import unittest
import numpy as np
from .test_utilities import *
import abs.sampler as cs
from abs import poisson_disk_downsample
import abs.utils as au


class TestSampling(unittest.TestCase):
    def test_sample_2dcurve(self):
        curve = circle2d()
        cs._uniform_sample_curve(curve, 2000)
        cs._random_sample_curve(curve, 2000)

    def test_sample_3dcurve(self):
        curve = ellipse3d()
        cs._uniform_sample_curve(curve, 2000)
        cs._random_sample_curve(curve, 2000)

    def test_sample_bspline_curve2d(self):
        curve = bspline_curve2d()
        cs._uniform_sample_curve(curve, 2000)
        cs._random_sample_curve(curve, 2000)

    def test_downsample(self):
        pts = np.random.rand(1000, 3)
        indices = poisson_disk_downsample(pts, 100)

    #TODO: Fix this test
    # def test_downsample1(self):
    #     pp = plane()
    #     _, pts = cs._random_sample_curve(pp, 5000)
    #     indices = poisson_disk_downsample(pts, 1000, 50)
    #     au.save_obj('test_sampling.obj', pts[indices])


if __name__ == '__main__':
    unittest.main()



"""
Combines geometry and topology to form complete shapes (parts).
Provides the Shape class that assembles curves, surfaces, and topology into a structured object.
"""
from . import sampler
from .topology import Edge, Face, Halfedge, Loop, Shell, Solid as TopoSolid, Topology
from .curve import create_curve
from .surface import create_surface
from .winding_number import find_surface_uv_for_curve
import numpy as np






class Shape:
    """Represents a geometric part (one part from the CAD model) with geometry and topology assembled."""
    def __init__(self, geometry_data, topology_data, spacing=0.02):

        self._geometry_data = self._geometry_data(geometry_data)
        self._topology_data = self._topology_data(topology_data)
        spacing *= np.linalg.norm(self._geometry_data.bbox[0][1] - self._geometry_data.bbox[0][0])
        self.bbox = self._geometry_data.bbox
        self.vertices = self._geometry_data.vertices

        self._create_2d_trimming_curves(self._geometry_data.curves2d, self._geometry_data.curves3d, spacing)
        self.Solid = self.Solid(self._topology_data, self._geometry_data, self.trimming_curves_2d)


    def _create_2d_trimming_curves(self, curves_2d, curves_3d, spacing):
        """
        Create 2D trimming curves.
        """
        self.trimming_curves_2d = []
        for shell in self._topology_data.shells:
            self._process_2d_trimming_curves_for_shell(shell, curves_2d, curves_3d, spacing)


    def _process_2d_trimming_curves_for_shell(self, shell_index, curves2d, curves3d, spacing):

        if isinstance(shell_index, Shell):
            shell = shell_index
        else:
            shell = self._topology_data.shells[shell_index]

        for (face_index, _) in shell.faces:
            face = self._topology_data.faces[face_index]
            self.trimming_curves_2d += (face_index-len(self.trimming_curves_2d)+1)* [None]
            self.trimming_curves_2d[face_index] = []

            surface_index = face.surface
            surface = self._geometry_data.surfaces[surface_index]

            for loop_id in face.loops:
                loop = self._topology_data.loops[loop_id]
                for halfedge_index in loop.halfedges:
                    halfedge = self._topology_data.halfedges[halfedge_index]
                    modified_orientation = halfedge.orientation_wrt_edge
                    if not face.surface_orientation:
                        modified_orientation = not modified_orientation

                    curve3d_index = halfedge.edge
                    curve3d = curves3d[curve3d_index]

                    if hasattr(halfedge, 'curve2d'):
                        curve2d_index = halfedge.curve2d
                        current_curve = curves2d[curve2d_index]
                    elif curve3d is not None:
                        current_curve = curve3d
                    else:
                        continue

                    if current_curve.shape_name == 'Line':
                        n_samples = 2
                    else:
                        length = current_curve.get_length()
                        n_samples = int(length / spacing)

                    if hasattr(halfedge, 'curve2d'):
                        curve2d_index = halfedge.curve2d
                        curve2d = curves2d[curve2d_index]
                        _, closest_surface_uv_values_of_curve = sampler.uniform_sample(curve2d, n_samples, 4, 100)
                        if not modified_orientation:
                            closest_surface_uv_values_of_curve = closest_surface_uv_values_of_curve[::-1]
                    else:
                        surface_uv_values, surface_points = sampler.uniform_sample(surface, n_samples*n_samples, 5, 100)


                        # Sample the curve points to get UV values
                        _, curve_points = sampler.uniform_sample(curve3d, n_samples)

                        if not modified_orientation:
                            curve_points = curve_points[::-1]
                        # Calculate the nearest UV values on the surface for the curve points
                        closest_surface_uv_values_of_curve = find_surface_uv_for_curve(surface_points, surface_uv_values, curve_points)

                    self.trimming_curves_2d[face_index].append(closest_surface_uv_values_of_curve)




    class _geometry_data:
        def __init__(self, geometry_data):
            self.curves2d, self.curves3d, self.surfaces, self.bbox, self.vertices = [], [], [], [], []
            self.__init_geometry(geometry_data)

        def __init_geometry(self, data):
            tmp = data.get('2dcurves', {}).values()
            self.curves2d=len(tmp)*[None]
            for curve_data in tmp:
                index, curve = create_curve(curve_data)
                self.curves2d[index] = curve

            tmp = data.get('3dcurves', {}).values()
            self.curves3d=len(tmp)*[None]
            for curve_data in tmp:
                index, curve = create_curve(curve_data)
                self.curves3d[index] = curve

            tmp = data.get('surfaces', {}).values()
            self.surfaces=len(tmp)*[None]
            for surface_data in tmp:
                index, surface = create_surface(surface_data)
                self.surfaces[index] = surface

            self.bbox.append(np.array(data.get('bbox')[:]))

            self.vertices.append(np.array(data.get('vertices')))


    class _topology_data:
        def __init__(self, topology_data):
            self.edges, self.faces, self.halfedges, self.loops, self.shells, self.solids = [], [], [], [], [], []
            self.__init_topology(topology_data)

        def __init_topology(self, data):

            entity_map = {
                'edges': (self.edges, _get_edges),
                'faces': (self.faces, _get_faces),
                'halfedges': (self.halfedges, _get_halfedges),
                'loops': (self.loops, _get_loops),
                'shells': (self.shells, Shell),
                'solids': (self.solids, TopoSolid)
            }

            for entity, (attr_list, constructor) in entity_map.items():
                entity_data = Topology._get_topo_data(data, entity)
                attr_list.extend(constructor(item) for item in entity_data)


    class Solid:
        def __init__(self, topology, geometry, trimming_curves):
            self.edges, self.faces, self.halfedges, self.loops, self.shells, self.solids = [], [], [], [], [], []
            self.__init_solid(topology, geometry, trimming_curves)


        def __init_solid(self, topo, geo, trimming_curves):

            # Loop over edges
            for edge in topo.edges:
                edge.curve3d = geo.curves3d[edge.curve3d]
                self.edges.append(edge)

            # loop over halfedges
            for halfedge in topo.halfedges:
                halfedge.curve2d = geo.curves2d[halfedge.curve2d] if halfedge.curve2d < len(geo.curves2d) else None
                halfedge.edge = self.edges[halfedge.edge]
                if halfedge.mates:
                    halfedge.mates = topo.halfedges[int(halfedge.mates[0])]
                self.halfedges.append(halfedge)

            # Loop over loops
            for loop in topo.loops:
                loop.halfedges = [topo.halfedges[halfedge_id] for halfedge_id in loop.halfedges]
                self.loops.append(loop)

            # Loop over faces
            for idx, face in enumerate(topo.faces):
                face.surface = geo.surfaces[face.surface]
                face.loops = [topo.loops[loop_id] for loop_id in face.loops]
                face.trimming_curves_2d = trimming_curves[idx]
                self.faces.append(face)

            # Loop over shells
            for shell in topo.shells:
                for idx, orientation in enumerate(shell.faces):
                    shell.faces[idx] = (topo.faces[orientation[0]] , orientation[1])
                self.shells.append(shell)

            # loop over solids
            for solid in topo.solids:
                solid.shells = [topo.shells[shell_id] for shell_id in solid.shells]
                self.solids.append(solid)

            # adding the reverse mapping

            # from edges to halfedges
            edgeMap = {}
            for halfEdge in self.halfedges:
                edge = halfEdge.edge
                edgeMapValue = edgeMap.get(edge, {'halfedges': []})
                edgeMapValue['halfedges'].append(halfEdge)
                edgeMap[edge] = edgeMapValue

            for edge in edgeMap:
                edge._halfedges = edgeMap[edge]['halfedges']

            # from halfedges to loops
            halfEdgeMap = {}
            for loop in self.loops:
                for halfedge in loop.halfedges:
                    halfEdgeMapValue = halfEdgeMap.get(halfedge, {'loops': []})
                    halfEdgeMapValue['loops'].append(loop)
                    halfEdgeMap[halfedge] = halfEdgeMapValue

            for halfedge in halfEdgeMap:
                halfedge.loops = halfEdgeMap[halfedge]['loops']

            # from loops to faces
            loopMap = {}
            for face in self.faces:
                for loop in face.loops:
                    loopMapValue = loopMap.get(loop, {'faces': []})
                    loopMapValue['faces'].append(face)
                    loopMap[loop] = loopMapValue

            for loop in loopMap:
                loop.faces = loopMap[loop]['faces']

            # from faces to shells
            faceMap = {}
            for shell in self.shells:
                for face, _ in shell.faces:
                    faceMapValue = faceMap.get(face, {'shells': []})
                    faceMapValue['shells'].append(shell)
                    faceMap[face] = faceMapValue

            for face in faceMap:
                face.shells = faceMap[face]['shells']

            # from shells to solids
            shellMap = {}
            if len(self.solids) > 0:
                for solid in self.solids:
                    for shell in solid.shells:
                        shellMapValue = shellMap.get(shell, {'solids': []})
                        shellMapValue['solids'].append(solid)
                        shellMap[shell] = shellMapValue

                for shell in shellMap:
                    shell.solids = shellMap[shell]['solids']





# Helper functions to construct topology objects
def _get_edges(edge_data): return Edge(edge_data)
def _get_faces(face_data): return Face(face_data)
def _get_halfedges(halfedge_data): return Halfedge(halfedge_data)
def _get_loops(loop_data): return Loop(loop_data)



import os
from pathlib import Path
import h5py
from test_geometry import *
from abs import *
import unittest
import numpy as np
from abs.utils import *

class Hdf5test(unittest.TestCase):

    def test_geometry_parts(self):
        file_path = get_file('Cone.hdf5')
        data = h5py.File(file_path, 'r')
        self.assertIsNotNone(data)
        self.assertIsNotNone(data['parts']['part_001']['geometry'])

    #  2D curves
    def test_line2d(self):
        sample_name = 'Cone.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/2dcurves/000']
            line = Line(grp)

        self.assertEqual(line.location.shape, (1, 2))
        self.assertEqual(line.interval.shape, (1, 2))
        self.assertEqual(line.direction.shape, (1, 2))

        sample_points = np.linspace(0, 1, 10).reshape(-1, 1)

        # sampling
        self.assertEqual(line.sample(sample_points).shape, (10, 2))

        # derivative
        self.assertEqual(line.derivative(sample_points, 0).shape, (10, 2))
        self.assertEqual(line.derivative(sample_points, 1).shape, (10, 2))
        self.assertEqual(line.derivative(sample_points, 2).shape, (10, 2))

        d, d2 = curves_derivative(line, sample_points)
        self.assertTrue(d < 1e-7)
        self.assertTrue(d2 < 1e-7)

        # length
        num_samples = 1000  # Can be adjusted for precision
        param_points, points = generate_points_on_curve(line, num_samples)
        self.assertTrue(abs(np.sum(np.linalg.norm(np.diff(points, axis=0), axis=1)) - line.get_length() < 1e-4))

        # normals
        # rotated_p = estimate_normal(line, num_samples)
        # self.assertTrue(abs(np.sum(rotated_p - line.normal(param_points)[1:, :]) < 1e-4))

        # check if normals are unit length
        # self.assertTrue(np.allclose(np.linalg.norm(line.normal(param_points), axis=1), 1, atol=1e-8))

    def test_circle2d(self):
        sample_name = 'Cylinder_Hole.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/2dcurves/004']
            circle = Circle(grp)

        sample_points = np.linspace(0, 2 * np.pi, 10).reshape(-1, 1)
        self.assertEqual(circle.sample(sample_points).shape, (10, 2))

        self.assertEqual(circle.location.shape, (1, 2))
        self.assertEqual(type(circle.radius), float)
        self.assertEqual(circle.interval.shape, (1, 2))
        self.assertEqual(circle.x_axis.shape, (1, 2))
        self.assertEqual(circle.y_axis.shape, (1, 2))

        # derivative
        self.assertEqual(circle.derivative(sample_points, 0).shape, (10, 2))
        self.assertEqual(circle.derivative(sample_points, 1).shape, (10, 2))
        self.assertEqual(circle.derivative(sample_points, 2).shape, (10, 2))
        self.assertEqual(circle.derivative(sample_points, 3).shape, (10, 2))
        self.assertEqual(circle.derivative(sample_points, 4).shape, (10, 2))

        d, d2 = curves_derivative(circle, sample_points)
        self.assertTrue(d < 1e-4)
        self.assertTrue(d2 < 1e-4)

        # length
        num_samples = 1000  # Can be adjusted for precision
        param_points, points = generate_points_on_curve(circle, num_samples)
        self.assertTrue(abs(np.sum(np.linalg.norm(np.diff(points, axis=0), axis=1)) - circle.get_length() < 1e-4))

        # normals
        # rotated_p = estimate_normal(circle, num_samples)
        # self.assertTrue(abs(np.sum(rotated_p - circle.normal(param_points)[1:, :]) < 1e-4))

        # check if normals are unit length
        # self.assertTrue(np.allclose(np.linalg.norm(circle.normal(param_points), axis=1), 1, atol=1e-8))

    def test_bspline_2dcurve(self):

        sample_name = 'SingleSolidSphere.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/2dcurves/001']
            bspline_curve2d = BSplineCurve(grp)

        self.assertEqual(bspline_curve2d.closed, False)
        self.assertEqual(type(bspline_curve2d.continuity), int)
        self.assertEqual(type(bspline_curve2d.degree), int)
        self.assertEqual(bspline_curve2d.interval.shape, (1, 2))
        self.assertEqual(bspline_curve2d.knots.shape[0], 1)
        self.assertEqual(bspline_curve2d.poles.shape[1], 2)
        self.assertEqual(type(bspline_curve2d.rational), bool)
        self.assertEqual(bspline_curve2d.weights.shape[1], 1)
        self.assertEqual(bspline_curve2d.poles.shape[0], bspline_curve2d.weights.shape[0])

        # sample points
        umin_value, umax_value = bspline_curve2d.interval.T
        gridX = np.linspace(umin_value, umax_value)
        sample_points = gridX.reshape(-1, 1)
        self.assertEqual(bspline_curve2d.sample(sample_points).shape, (gridX.shape[0], 2))

        self.assertEqual(bspline_curve2d.derivative(sample_points, 0).shape, (50, 2))
        self.assertEqual(bspline_curve2d.derivative(sample_points, 1).shape, (50, 2))
        d, d2 = curves_derivative(bspline_curve2d, sample_points)
        self.assertTrue(d < 1e-4)
        self.assertTrue(d2 < 1e-4)

        # length
        num_samples = 1000  # Can be adjusted for precision
        param_points, points = generate_points_on_curve(bspline_curve2d, num_samples)
        self.assertTrue(abs(np.sum(np.linalg.norm(np.diff(points, axis=0), axis=1)) - bspline_curve2d.get_length() < 1e-4))

    # 3D curves
    def test_line3d(self):
        sample_name = 'Box.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/3dcurves/000']
            line = Line(grp)

        self.assertEqual(line.location.shape, (1, 3))
        self.assertEqual(line.direction.shape, (1, 3))
        self.assertEqual(line.interval.shape, (1, 2))
        sample_points: None = np.linspace(0, 1, 10).reshape(-1, 1)

        # sampling
        self.assertEqual(line.sample(sample_points).shape, (10, 3))

        # derivative shape
        self.assertEqual(line.derivative(sample_points, 0).shape, (10, 3))
        self.assertEqual(line.derivative(sample_points, 1).shape, (10, 3))
        self.assertEqual(line.derivative(sample_points, 2).shape, (10, 3))

        d, d2 = curves_derivative(line, sample_points)
        self.assertTrue(d < 1e-7)
        self.assertTrue(d2 < 1e-7)


    def test_circle3d(self):
        sample_name = 'Cone.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/3dcurves/000']
            circle = Circle(grp)

        self.assertEqual(circle.location.shape, (1, 3))
        self.assertEqual(type(circle.radius), float)
        self.assertEqual(circle.interval.shape, (1, 2))
        self.assertEqual(circle.x_axis.shape, (1, 3))
        self.assertEqual(circle.y_axis.shape, (1, 3))
        self.assertEqual(circle.z_axis.shape, (1, 3))
        sample_points = np.linspace(0, 2 * np.pi, 10).reshape(-1, 1)
        self.assertEqual(circle.sample(sample_points).shape, (10, 3))

        # derivative
        self.assertEqual(circle.derivative(sample_points, 0).shape, (10, 3))
        self.assertEqual(circle.derivative(sample_points, 1).shape, (10, 3))
        self.assertEqual(circle.derivative(sample_points, 2).shape, (10, 3))
        self.assertEqual(circle.derivative(sample_points, 3).shape, (10, 3))
        self.assertEqual(circle.derivative(sample_points, 4).shape, (10, 3))

        d, d2 = curves_derivative(circle, sample_points)
        self.assertTrue(d < 1e-4)
        self.assertTrue(d2 < 1e-4)

    def test_ellipse3d(self):
        sample_name = 'Ellipse.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/3dcurves/000']
            ellipse = Ellipse(grp)

        self.assertEqual(ellipse.focus1.shape, (1, 3))
        self.assertEqual(ellipse.focus2.shape, (1, 3))
        self.assertEqual(ellipse.interval.shape, (1, 2))
        self.assertEqual(type(ellipse.maj_radius), float)
        self.assertEqual(type(ellipse.min_radius), float)
        self.assertEqual(ellipse.x_axis.shape, (1, 3))
        self.assertEqual(ellipse.y_axis.shape, (1, 3))
        self.assertEqual(ellipse.z_axis.shape, (1, 3))

        sample_points = np.linspace(0, 2 * np.pi, 10).reshape(-1, 1)
        self.assertEqual(ellipse.sample(sample_points).shape, (10, 3))

        # derivative
        self.assertEqual(ellipse.derivative(sample_points, 0).shape, (10, 3))
        self.assertEqual(ellipse.derivative(sample_points, 1).shape, (10, 3))
        self.assertEqual(ellipse.derivative(sample_points, 2).shape, (10, 3))
        self.assertEqual(ellipse.derivative(sample_points, 3).shape, (10, 3))
        self.assertEqual(ellipse.derivative(sample_points, 4).shape, (10, 3))

        d, d2 = curves_derivative(ellipse, sample_points)
        self.assertTrue(d < 1e-4)
        self.assertTrue(d2 < 1e-4)

    def test_bspline_3dcurve(self):
        sample_name = 'Cone.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/3dcurves/001']
            bspline_curve3d = BSplineCurve(grp)

        self.assertEqual(type(bspline_curve3d.closed), bool)
        self.assertEqual(type(bspline_curve3d.continuity), int)
        self.assertEqual(type(bspline_curve3d.degree), int)
        self.assertEqual(bspline_curve3d.interval.shape, (1, 2))
        self.assertEqual(bspline_curve3d.knots.shape[0], 1)
        self.assertEqual(bspline_curve3d.poles.shape[1], 3)
        self.assertEqual(type(bspline_curve3d.rational), bool)
        self.assertEqual(bspline_curve3d.weights.shape[1], 1)

        # sample points
        umin_value, umax_value = bspline_curve3d.interval.T
        gridX = np.linspace(umin_value, umax_value)
        sample_points = gridX.reshape(-1, 1)
        self.assertEqual(bspline_curve3d.sample(sample_points).shape, (gridX.shape[0], 3))

        # derivative
        self.assertEqual(bspline_curve3d.derivative(sample_points, 0).shape, (50, 3))
        self.assertEqual(bspline_curve3d.derivative(sample_points, 1).shape, (50, 3))
        d, d2 = curves_derivative(bspline_curve3d, sample_points)
        self.assertTrue(d < 1e-4)
        self.assertTrue(d2 < 1e-4)


    # Surfaces
    def test_plane(self):
        sample_name = 'Box.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/surfaces/001']
            plane = Plane(grp)

        self.assertEqual(plane.coefficients.shape, (1, 4))
        self.assertEqual(plane.location.shape, (1, 3))
        self.assertEqual(plane.trim_domain.shape, (2, 2))
        self.assertEqual(plane.x_axis.shape, (1, 3))
        self.assertEqual(plane.y_axis.shape, (1, 3))
        self.assertEqual(plane.z_axis.shape, (1, 3))

        # sample points
        umin_value, umax_value, vmin_value, vmax_value = plane.trim_domain.reshape(-1, 1)
        gridX = np.linspace(umin_value, umax_value)
        gridY = np.linspace(vmin_value, vmax_value)
        gridX, gridY = np.meshgrid(gridX, gridY)
        gridX = gridX.reshape((np.prod(gridX.shape),)).reshape(-1, 1)
        gridY = gridY.reshape((np.prod(gridY.shape),)).reshape(-1, 1)
        sample_points = np.column_stack((gridX, gridY)).reshape(-1, 2)

        self.assertEqual(plane.sample(sample_points).shape, (gridX.shape[0] * gridX.shape[1], 3))

        # derivative:
        p, q, du, dv = surface_derivative(plane, sample_points)
        self.assertTrue(p < 1e-7)
        self.assertTrue(q < 1e-7)
        self.assertTrue(du < 1e-7)
        self.assertTrue(dv < 1e-7)

    def test_cylinder(self):
        sample_name = 'Cylinder_Hole.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/surfaces/000']
            cylinder = Cylinder(grp)

        self.assertEqual(cylinder.location.shape, (1, 3))
        self.assertEqual(type(cylinder.radius), float)
        self.assertEqual(cylinder.coefficients.shape, (1, 10))
        self.assertEqual(cylinder.trim_domain.shape, (2, 2))
        self.assertEqual(cylinder.x_axis.shape, (1, 3))
        self.assertEqual(cylinder.y_axis.shape, (1, 3))
        self.assertEqual(cylinder.z_axis.shape, (1, 3))

        # sample points
        umin_value, umax_value, vmin_value, vmax_value = cylinder.trim_domain.reshape(-1, 1)
        gridX = np.linspace(umin_value, umax_value)
        gridY = np.linspace(vmin_value, vmax_value)
        gridX, gridY = np.meshgrid(gridX, gridY)
        sample_points = np.column_stack((gridX, gridY)).reshape(-1, 2)
        self.assertEqual(cylinder.sample(sample_points).shape, (gridX.shape[0] * gridX.shape[1], 3))

        du, dv, d2u, d2v = surface_derivative(cylinder, sample_points)
        self.assertTrue(du < 1e-4)
        self.assertTrue(dv < 1e-4)
        self.assertTrue(d2u < 1e-4)
        self.assertTrue(d2v < 1e-4)


    def test_sphere(self):
        sample_name = 'Sphere.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/surfaces/000']
            sphere = Sphere(grp)

        self.assertEqual(sphere.coefficients.shape, (1, 10))
        self.assertEqual(sphere.location.shape, (1, 3))
        self.assertEqual(type(sphere.radius), float)
        self.assertEqual(sphere.trim_domain.shape, (2, 2))
        self.assertEqual(sphere.x_axis.shape, (1, 3))
        self.assertEqual(sphere.y_axis.shape, (1, 3))
        self.assertEqual(sphere.z_axis.shape, (1, 3))

        # sample points
        umin_value, umax_value, vmin_value, vmax_value = sphere.trim_domain.reshape(-1, 1)
        gridX = np.linspace(umin_value, umax_value)
        gridY = np.linspace(vmin_value, vmax_value)
        gridX, gridY = np.meshgrid(gridX, gridY)
        sample_points = np.column_stack((gridX, gridY)).reshape(-1, 2)
        self.assertEqual(sphere.sample(sample_points).shape, (gridX.shape[0] * gridX.shape[1], 3))

        # derivative: du, dv
        du, dv, d2u, d2v = surface_derivative(sphere, sample_points)
        self.assertTrue(du < 1e-4)
        self.assertTrue(dv < 1e-4)
        self.assertTrue(d2u < 1e-4)
        self.assertTrue(d2v < 1e-4)

    def test_torus(self):
        sample_name = 'Cylinder_Hole_Fillet_Chamfer.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/surfaces/004']
            torus = Torus(grp)

        self.assertEqual(torus.location.shape, (1, 3))
        self.assertEqual(type(torus.max_radius), float)
        self.assertEqual(type(torus.min_radius), float)
        self.assertEqual(torus.trim_domain.shape, (2, 2))
        self.assertEqual(torus.x_axis.shape, (1, 3))
        self.assertEqual(torus.y_axis.shape, (1, 3))
        self.assertEqual(torus.z_axis.shape, (1, 3))

        # sample points
        umin_value, umax_value, vmin_value, vmax_value = torus.trim_domain.reshape(-1, 1)
        gridX = np.linspace(umin_value, umax_value)
        gridY = np.linspace(vmin_value, vmax_value)
        gridX, gridY = np.meshgrid(gridX, gridY)
        sample_points = np.column_stack((gridX, gridY)).reshape(-1, 2)
        self.assertEqual(torus.sample(sample_points).shape, (gridX.shape[0] * gridX.shape[1], 3))

        du, dv, d2u, d2v = surface_derivative(torus, sample_points)
        self.assertTrue(du < 1e-4)
        self.assertTrue(dv < 1e-4)
        self.assertTrue(d2u < 1e-4)
        self.assertTrue(d2v < 1e-4)


    def test_bspline_surface(self):

        sample_name = 'Ellipse.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/surfaces/000']
            bspline_surface = BSplineSurface(grp)

        self.assertEqual(type(bspline_surface.continuity), int)
        self.assertEqual(bspline_surface.face_domain.shape, (1, 4))
        self.assertEqual(type(bspline_surface.is_trimmed), bool)
        self.assertEqual(bspline_surface.poles.shape[2], 3)
        self.assertEqual(bspline_surface.trim_domain.shape, (2, 2))
        self.assertEqual(type(bspline_surface.u_closed), bool)
        self.assertEqual(type(bspline_surface.u_degree), int)
        self.assertEqual(bspline_surface.u_knots.shape[0], 1)
        self.assertEqual(type(bspline_surface.u_rational), bool)
        self.assertEqual(type(bspline_surface.v_closed), bool)
        self.assertEqual(type(bspline_surface.v_degree), int)
        self.assertEqual(bspline_surface.v_knots.shape[0], 1)
        self.assertEqual(type(bspline_surface.v_rational), bool)

        # sample points
        umin_value, umax_value, vmin_value, vmax_value = bspline_surface.trim_domain.reshape(-1, 1)
        gridX = np.linspace(umin_value, umax_value)
        gridY = np.linspace(vmin_value, vmax_value)
        gridX, gridY = np.meshgrid(gridX, gridY)
        sample_points = np.column_stack((gridX, gridY)).reshape(-1, 2)
        self.assertEqual(bspline_surface.sample(sample_points).shape, (gridX.shape[0] * gridX.shape[1], 3))

        # derivative
        du, dv, d2u, d2v = surface_derivative(bspline_surface, sample_points, 1e-7)
        self.assertTrue(du < 1e-4)
        self.assertTrue(dv < 1e-4)
        self.assertTrue(d2u < 1e-4)
        self.assertTrue(d2v < 1e-4)

    def test_cone(self):
        sample_name = 'Cone.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/geometry/surfaces/000']
            cone = Cone(grp)

        self.assertEqual(type(cone.angle), float)
        self.assertEqual(cone.apex.shape, (1, 3))
        self.assertEqual(cone.coefficients.shape, (1, 10))
        self.assertEqual(cone.location.shape, (1, 3))
        self.assertEqual(type(cone.radius), float)
        self.assertEqual(cone.trim_domain.shape, (2, 2))
        self.assertEqual(cone.x_axis.shape, (1, 3))
        self.assertEqual(cone.y_axis.shape, (1, 3))
        self.assertEqual(cone.z_axis.shape, (1, 3))

        # sample points
        umin_value, umax_value, vmin_value, vmax_value = cone.trim_domain.reshape(-1, 1)
        gridX = np.linspace(umin_value, umax_value)
        gridY = np.linspace(vmin_value, vmax_value)
        gridX, gridY = np.meshgrid(gridX, gridY)
        sample_points = np.column_stack((gridX, gridY)).reshape(-1, 2)
        self.assertEqual(cone.sample(sample_points).shape, (gridX.shape[0] * gridX.shape[1], 3))

        du, dv, d2u, d2v = surface_derivative(cone, sample_points)
        self.assertTrue(du < 1e-4)
        self.assertTrue(dv < 1e-4)
        self.assertTrue(d2u < 1e-4)
        self.assertTrue(d2v < 1e-4)

if __name__ == '__main__':
    unittest.main()



"""
Curve classes for parametric curves (Line, Circle, Ellipse, B-spline, etc).
Provide methods to sample points on the curve and compute derivatives.
"""
import numpy as np
from scipy.interpolate import BSpline
from geomdl import NURBS


def create_curve(curve_data, compute_index=True):
    """
    Factory function to create a Curve object from HDF5 curve data.
    Returns (index, curve_object). If compute_index is False, index is None.
    """
    if compute_index:
        index = int(curve_data.name.split("/")[-1])
    else:
        index = None
    curve_type = curve_data.get('type')[()].decode('utf-8')
    curve_map = {
        'Line': Line,
        'Circle': Circle,
        'Ellipse': Ellipse,
        'BSpline': BSplineCurve,
        'Other': Other
    }
    curve_class = curve_map.get(curve_type)
    if curve_class:
        return index, curve_class(curve_data)
    else:
        # Unsupported curve type
        return index, None


class Curve:
    """Base class for all curve types."""
    def sample(self, points):
        """Sample points on the curve at given parametric values."""
        raise NotImplementedError("Sample method must be implemented by subclasses")
    def derivative(self, points, order=1):
        """Evaluate the curve's derivative of given order at parametric points."""
        raise NotImplementedError("Derivative method must be implemented by subclasses")
    def get_length(self):
        """Approximate the curve length via Gaussian quadrature (4-point Legendre)."""
        if getattr(self, "length", None) is not None and self.length != -1:
            return self.length
        # Compute length via integration of first derivative
        pts, weights = np.polynomial.legendre.leggauss(4)
        pts = (pts + 1) * 0.5 * (self.interval[0, 1] - self.interval[0, 0]) + self.interval[0, 0]
        deriv = self.derivative(pts[:, None], order=1)
        lengths = np.linalg.norm(deriv, axis=1)
        self.length = np.sum(lengths * weights) * (self.interval[0, 1] - self.interval[0, 0]) / 2
        return self.length



class Line(Curve):
    """Straight line curve defined by a start location and direction vector."""
    def __init__(self, line):
        self.location = np.array(line.get('location')[()]).reshape(-1, 1).T
        self.interval = np.array(line.get('interval')[()]).reshape(-1, 1).T
        self.direction = np.array(line.get('direction')[()]).reshape(-1, 1).T
        if self.direction.shape[1] == 3:
            self.transform = np.array(line.get('transform')[()])  # transformation matrix if present
        self.length = -1
        self.shape_name = line.get('type')[()].decode('utf8')

    def sample(self, sample_points):
        if sample_points.size == 0:
            return self.location
        # Parametric coordinate in interval [0,1] maps to location + t * direction
        return self.location + sample_points * self.direction

    def get_length(self):
        if self.length == -1:
            self.length = np.linalg.norm((self.interval[0, 1] - self.interval[0, 0]) * self.direction)
        return self.length

    def derivative(self, sample_points, order=1):
        if order == 1:
            # First derivative of line is constant direction
            return np.tile(self.direction, (sample_points.shape[0], 1))
        # Higher derivatives of a line are zero
        return np.zeros((sample_points.shape[0], self.location.shape[1]))


class Circle(Curve):
    """Circular arc or full circle curve."""
    def __init__(self, circle):
        self.location = np.array(circle.get('location')[()]).reshape(-1, 1).T
        self.radius = float(circle.get('radius')[()])
        self.interval = np.array(circle.get('interval')[()]).reshape(-1, 1).T
        self.x_axis = np.array(circle.get('x_axis')[()]).reshape(-1, 1).T
        self.y_axis = np.array(circle.get('y_axis')[()]).reshape(-1, 1).T
        if 'z_axis' in circle:
            self.z_axis = np.array(circle.get('z_axis')[()]).reshape(-1, 1).T
            self.transform = np.array(circle.get('transform')[()])
        self.length = -1
        self.shape_name = circle.get('type')[()].decode('utf8')
    def sample(self, sample_points):
        if sample_points.size == 0:
            return self.location
        # parametric angle values in sample_points
        circle_points = self.location + self.radius * (np.cos(sample_points) * self.x_axis + np.sin(sample_points) * self.y_axis)
        return circle_points
    def derivative(self, sample_points, order=1):
        if order == 0:
            return self.sample(sample_points)
        # Circle derivatives are periodic every 4th derivative
        if order % 4 == 0 and order != 0:
            # Cosine form (same as first derivative for order 4)
            return self.radius * (np.cos(sample_points) * self.x_axis + np.sin(sample_points) * self.y_axis)
        elif order % 4 == 1:
            return -self.radius * np.sin(sample_points) * self.x_axis + self.radius * np.cos(sample_points) * self.y_axis
        elif order % 4 == 2:
            return -self.radius * (np.cos(sample_points) * self.x_axis + np.sin(sample_points) * self.y_axis)
        else:  # order % 4 == 3
            return self.radius * (np.sin(sample_points) * self.x_axis - np.cos(sample_points) * self.y_axis)


class Ellipse(Curve):
    """Elliptical curve (full ellipse or arc)."""
    def __init__(self, ellipse):
        self.focus1 = np.array(ellipse.get('focus1')[()]).reshape(-1, 1).T
        self.focus2 = np.array(ellipse.get('focus2')[()]).reshape(-1, 1).T
        self.interval = np.array(ellipse.get('interval')[()]).reshape(-1, 1).T
        self.maj_radius = float(ellipse.get('maj_radius')[()])
        self.min_radius = float(ellipse.get('min_radius')[()])
        self.x_axis = np.array(ellipse.get('x_axis')[()]).reshape(-1, 1).T
        self.y_axis = np.array(ellipse.get('y_axis')[()]).reshape(-1, 1).T
        self.length = -1
        self.shape_name = ellipse.get('type')[()].decode('utf8')
        if 'z_axis' in ellipse:
            self.z_axis = np.array(ellipse.get('z_axis')[()]).reshape(-1, 1).T
            self.transform = np.array(ellipse.get('transform')[()])  # 4x4 transform matrix if present
        # Center of ellipse
        self.center = (self.focus1 + self.focus2) / 2
    def sample(self, sample_points):
        # Parametric angle values in sample_points
        ellipse_points = self.center + self.maj_radius * np.cos(sample_points) * self.x_axis + self.min_radius * np.sin(sample_points) * self.y_axis
        return ellipse_points
    def derivative(self, sample_points, order=1):
        if order % 4 == 0:
            if order == 0:
                return self.sample(sample_points)
            return self.maj_radius * np.cos(sample_points) * self.x_axis + self.min_radius * np.sin(sample_points) * self.y_axis
        elif order % 4 == 1:
            return -self.maj_radius * np.sin(sample_points) * self.x_axis + self.min_radius * np.cos(sample_points) * self.y_axis
        elif order % 4 == 2:
            return -self.maj_radius * np.cos(sample_points) * self.x_axis - self.min_radius * np.sin(sample_points) * self.y_axis
        else:
            return self.maj_radius * np.sin(sample_points) * self.x_axis - self.min_radius * np.cos(sample_points) * self.y_axis


# class BSplineCurve(Curve):
#     """B-spline or NURBS curve."""
#     def __init__(self, bspline):
#         self.length = -1
#         self.closed = bool(bspline.get('closed')[()])
#         self.degree = int(bspline.get('degree')[()])
#         self.continuity = int(bspline.get('continuity')[()])
#         self.poles = np.array(bspline.get('poles')[()])  # control points array
#         self.knots = np.array(bspline.get('knots')[()]).reshape(-1, 1).T
#         self.weights = np.array(bspline.get('weights')[()]).reshape(-1, 1)
#         self.interval = np.array(bspline.get('interval')[()]).reshape(-1, 1).T
#         self.rational = bool(bspline.get('rational')[()])
#         self.periodic = bool(bspline.get('periodic')[()])
#         self.shape_name = bspline.get('type')[()].decode('utf8')
#         if self.poles.shape[1] == 3:
#             self.transform = np.array(bspline.get('transform')[()])  # apply if present
#         # Create underlying BSpline or NURBS curve object
#         if self.rational:
#             # Use NURBS-Python for rational B-spline
#             self.bspline = NURBS.Curve()  # no normalization of knots to allow exact usage
#             self.bspline.degree = self.degree
#             self.bspline.ctrlpts = self.poles.tolist()
#             self.bspline.knotvector = self.knots.flatten().tolist()
#             self.bspline.weights = self.weights.flatten().tolist()
#         else:
#             # Use SciPy BSpline for non-rational
#             self.bspline = BSpline(self.knots.T[0], self.poles, self.degree)
#     def sample(self, sample_points):
#         if sample_points.size == 0:
#             return np.array([]).reshape(0, self.poles.shape[1])
#         if self.rational:
#             # Evaluate rational B-spline (list of points)
#             return np.array(self.bspline.evaluate_list(sample_points.flatten().tolist()))
#         # Evaluate SciPy BSpline (returns (n_points, dim) array)
#         return np.squeeze(self.bspline(sample_points))
#     def derivative(self, sample_points, order=1):
#         if order == 0:
#             return self.sample(sample_points)
#         # If requested derivative order is higher than degree, derivative is zero vector
#         if self.degree < order:
#             return np.zeros((sample_points.shape[0], self.poles.shape[1]))
#         if self.rational:
#             # Use geomdl to compute derivatives for rational curves
#             res = np.zeros((sample_points.shape[0], self.poles.shape[1]))
#             for i in range(sample_points.shape[0]):
#                 d = self.bspline.derivatives(sample_points[i, 0], order)
#                 res[i, :] = d[-1]  # d[-1] is highest derivative (order-th)
#             return res
#         # For non-rational, SciPy's BSpline object can provide derivative by constructing a new BSpline
#         b_spline_deriv = self.bspline.derivative(nu=order)
#         return np.squeeze(b_spline_deriv(sample_points))
class BSplineCurve(Curve):
    def __init__(self, bspline):

        self.length = -1

        self.closed = bool(bspline.get('closed')[()])
        self.degree = int(bspline.get('degree')[()])
        self.continuity = int(bspline.get('continuity')[()])
        self.poles = np.array(bspline.get('poles')[()])
        self.knots = np.array(bspline.get('knots')[()]).reshape(-1, 1).T
        self.weights = np.array(bspline.get('weights')[()]).reshape(-1, 1)
        self.interval = np.array(bspline.get('interval')[()]).reshape(-1, 1).T
        self.rational = bool(bspline.get('rational')[()])
        self.periodic = bool(bspline.get('periodic')[()])
        self.shape_name = bspline.get('type')[()].decode('utf8')
        if self.poles.shape[1] == 3:
            self.transform = np.array(bspline.get('transform')[()])

        # Create BSpline or NURBS curve object


        if self.rational:
            self.bspline = NURBS.Curve(normalize_kv=False)
            self.bspline.degree = self.degree
            self.bspline.ctrlpts = self.poles.tolist()
            self.bspline.knotvector = self.knots.flatten().tolist()
            self.bspline.weights = self.weights.flatten().tolist()
        else:
            self.bspline = BSpline(self.knots.T[:,0], self.poles, self.degree)



    def sample(self, sample_points):
        if self.rational:
            return np.array(self.bspline.evaluate_list(sample_points.flatten().tolist()))

        return np.squeeze(self.bspline(sample_points))

    def derivative(self, sample_points, order=1):

        if order == 0:
            return self.sample(sample_points)

        elif self.degree < order:
            return np.zeros([sample_points.shape[0], self.poles.shape[1]])
        else:
            if self.rational:
                res = np.zeros([sample_points.shape[0], self.poles.shape[1]])
                for i in range(sample_points.shape[0]):
                    d = self.bspline.derivatives(sample_points[i, 0], order)
                    res[i, :] = d[-1]
                return res
            b_spline_derivative = self.bspline.derivative(order)
            return np.squeeze(b_spline_derivative(sample_points))
class Other(Curve):
    """Fallback for unsupported curve types."""
    def __init__(self, other):
        self.shape_name = other.get('type')[()].decode('utf8')
        self.interval = np.array(other.get('interval')[()]).reshape(-1, 1).T
    def sample(self, sample_points):
        # Unsupported: return empty array
        return np.array([]).reshape(0, 3)
    def derivative(self, sample_points, order=1):
        return np.array([]).reshape(0, 3)
    def get_length(self):
        return 0.0





"""
Sampling routines for curves and surfaces.
Provides uniform and random sampling of points along parametric entities.
"""

import numpy as np
from .curve import Curve
from .surface import Surface


def uniform_sample(topo, num_samples, min_pts=None, max_pts=None):
    """
    Sample approximately uniform distributed points on a curve or surface.

    Args:
    curve: The curve entity to be sampled.
    spacing (float): The spacing parameter for sampling.
    min_pts (int): Minimum number of points to sample.
    max_pts (int): Maximum number of points to sample.


    Returns:
    parametric values and an array of sampled points on the curve/surface.
    """

    if isinstance(topo, Curve):
        return _uniform_sample_curve(topo, num_samples, min_pts, max_pts)
    elif isinstance(topo, Surface):
        return _uniform_sample_surface(topo, num_samples, min_pts, max_pts)
    elif hasattr(topo, 'curve3d') and isinstance(topo.curve3d, Curve):
        return _uniform_sample_curve(topo.curve3d, num_samples, min_pts, max_pts)
    elif hasattr(topo, 'surface') and isinstance(topo.surface, Surface):
        return _uniform_sample_surface(topo.surface, num_samples, min_pts, max_pts)
    else:
        raise ValueError("Invalid geometry type")

def random_sample(topo, num_samples, min_pts=None, max_pts=None):
    """
    Sample random (uniformly random in parameter domain) points on a curve or surface.

    Args:
    curve: The curve entity to be sampled.
    spacing (float): The spacing parameter for sampling.
    min_pts (int): Minimum number of points to sample.
    max_pts (int): Maximum number of points to sample.


    Returns:
    parametric values and an array of sampled points on the curve/surface.
    """
    if isinstance(topo, Curve):
        return _random_sample_curve(topo, num_samples, min_pts, max_pts)
    elif isinstance(topo, Surface):
        return _random_sample_surface(topo, num_samples, min_pts, max_pts)
    elif hasattr(topo, 'curve3d') and isinstance(topo.curve3d, Curve):
        return _random_sample_curve(topo.curve3d, num_samples, min_pts, max_pts)
    elif hasattr(topo, 'surface') and isinstance(topo.surface, Surface):
        return _random_sample_surface(topo.surface, num_samples, min_pts, max_pts)
    else:
        raise ValueError("Invalid geometry type")



def _uniform_sample_curve(curve, num_samples, min_pts=None, max_pts=None):
    """
    Helper: Uniformly sample points on a Curve.

    Args:
    curve: The curve entity to be sampled.
    spacing (float): The spacing parameter for sampling.


    Returns:
    parametric values and an array of sampled points on the curve.
    """

    # num_samples = max(int(curve.length() / spacing), 1)

    if min_pts is not None:
        num_samples = max(num_samples, min_pts)
    if max_pts is not None:
        num_samples = min(num_samples, max_pts)
    # parameter values evenly spaced over interval
    t = np.linspace(curve.interval[0, 0], curve.interval[0, 1], num_samples).reshape(-1, 1)
    return t, curve.sample(t)


def _random_sample_curve(curve, num_samples, min_pts=None, max_pts=None):
    """
    Helper: Randomly sample points on a Curve.

    Args:
    curve: The curve entity to be sampled.
    spacing (float): The spacing parameter for sampling.


    Returns:
    parametric values and an array of sampled points on the curve.
    """

    if min_pts is not None:
        num_samples = max(num_samples, min_pts)
    if max_pts is not None:
        num_samples = min(num_samples, max_pts)
    # random parameters uniformly distributed in interval
    t = np.random.uniform(low=curve.interval[0, 0], high=curve.interval[0, 1], size=num_samples).reshape(-1, 1)
    return t, curve.sample(t)


def _uniform_sample_surface(surface, num_samples, min_pts=None, max_pts=None):
    """
    Helper: Uniformly sample points on a Surface (grid sampling).

    Args:
    surface: The surface entity to be sampled.
    spacing (float): The spacing parameter for sampling.


    Returns:
    parametric values and an array of sampled points on the surface.
    """

    if min_pts is not None:
        num_samples = max(num_samples, min_pts)
    if max_pts is not None:
        num_samples = min(num_samples, max_pts)

    num_samples = np.sqrt(num_samples).astype(int)
    # Parameter ranges
    u_values = np.linspace(surface.trim_domain[0, 0], surface.trim_domain[0, 1], num_samples)
    v_values = np.linspace(surface.trim_domain[1, 0], surface.trim_domain[1, 1], num_samples)

    points = np.array(np.meshgrid(u_values, v_values)).T.reshape(-1, 2)
    return points, surface.sample(points)


def _random_sample_surface(surface, num_samples, min_pts=None, max_pts=None):
    """
    Helper: Randomly sample points on a Surface.

    Args:
    surface: The surface entity to be sampled.
    spacing (float): The spacing parameter for sampling.


    Returns:
    parametric values and an array of sampled points on the surface.
    """

    if min_pts is not None:
        num_samples = max(num_samples, min_pts)
    if max_pts is not None:
        num_samples = min(num_samples, max_pts)
    # Uniform random sampling in the UV domain rectangle
    points = np.random.uniform(low=[surface.trim_domain[0, 0], surface.trim_domain[1, 0]],
                          high=[surface.trim_domain[0, 1], surface.trim_domain[1, 1]],
                          size=(num_samples,2))
    return points, surface.sample(points)




import numpy as np
import h5py
from typing import Any, Dict, List
from abs.winding_number import winding_number


class Topology:
    """Base class for topology entities (Edge, Face, etc.)"""
    @staticmethod
    def _get_topo_data(topo_data: h5py.File, entity: str) -> List[Dict[str, Any]]:

        def recursive_hdf5_to_dict(group: h5py.Group) -> Dict[str, Any]:

            return {
                key: recursive_hdf5_to_dict(item) if isinstance(item, h5py.Group) else item[()]
                for key, item in group.items()
            }

        if isinstance(topo_data[entity], h5py.Dataset):
            return []

        sorted_data = sorted(
            ((int(k), recursive_hdf5_to_dict(topo_data[entity][k])) for k in topo_data[entity].keys()),
            key=lambda x: x[0]
        )

        return [{**data,  **{"id": id}} for id, data in sorted_data]

    def is_face(self):
        return True if isinstance(self, Face) else False

    def is_edge(self):
        return True if isinstance(self, Edge) else False




class Edge(Topology):
    """Edge in the topology, with pointers to its 3D curve and vertices."""
    def __init__(self, edge):
        self.curve3d = edge['3dcurve']
        self.end_vertex = int(edge['end_vertex'])
        self.start_vertex = int(edge['start_vertex'])
        self.half_edges = None
        self.id = edge['id']


    def get_length(self):
        if self.curve3d:
            return self.curve3d.get_length()
        else:
            return 0

    def derivative(self, points, order=1):
        return self.curve3d.derivative(points, order)

    def sample(self, sample_points):
        return self.curve3d.sample(sample_points)



class Face(Topology):
    """Face in the topology, bounded by loops and referencing a surface."""
    def __init__(self, face):
        self.loops = np.array(face['loops'])
        self.surface = np.int64(face['surface'])
        self.surface_orientation = face['surface_orientation']
        self.trimming_curves_2d = []
        self.shells = None
        self.exact_domain = np.array(face['exact_domain'])
        self.nr_singularities = face['has_singularities']
        self.nr_singularities = face['nr_singularities']
        self.outer_loop = face['outer_loop']
        self.singularities = face['singularities']
        self.id = face['id']


    def normal(self, points):

        if self.surface.shape_name == 'Other':
            return None

        normal_points = self.surface.normal(points)
        if not self.surface_orientation:
            normal_points = -normal_points
        return normal_points

    def get_area(self):
        if self.surface is not None:
            return self.surface.get_area()
        else:
            return 0

    def derivative(self, points, order=1):
        return self.surface.derivative(points, order)

    def sample(self, sample_points):
        return self.surface.sample(sample_points)

    def find_adjacent_faces(self):
        """
        Find the adjacent faces of a face.
        """
        #TODO: fix this!
        adjacent_faces = set()
        for loop in self.loops:
            for halfedge in loop.halfedges:
                if halfedge.mates:
                    # is this correct?
                    for other_loop in halfedge.mates.loops:
                        for other_face in other_loop.faces:
                            adjacent_faces.add(other_face)
        adjacent_faces -= {self}
        return adjacent_faces


    def filter_outside_points(self, uv_points):
        """
        Filter out points that are outside the trimming curve of a face.
        """
        total_winding_numbers = np.zeros((len(uv_points), 1))
        curves = self.trimming_curves_2d
        for poly in curves:
            total_winding_numbers += winding_number(poly, uv_points)

        res = total_winding_numbers > 0.5
        res = res.reshape(-1)
        return res


class Halfedge(Topology):
    """Half-edge: directed edge segment, part of a loop boundary."""
    def __init__(self, halfedge):
        self.curve2d = np.int64(halfedge['2dcurve'])
        self.edge = np.int64(halfedge['edge'])
        self.orientation_wrt_edge = halfedge['orientation_wrt_edge']
        if halfedge['mates']:
            self.mates = np.array(halfedge['mates'])
        else:
            self.mates = None
        self.loops = None
        self.id = halfedge['id']


    def get_length(self):
        return self.curve2d.get_length()

    def derivative(self, points, order=1):
        return self.curve2d.derivative(points, order)

    def sample(self, sample_points):
        return self.curve2d.sample(sample_points)




class Loop(Topology):
    """Loop: a closed sequence of halfedges bounding a face."""
    def __init__(self, loop):
        self.halfedges = np.array(loop['halfedges'])
        self.faces = None

        self.id = loop['id']

class Shell(Topology):
    """Shell: collection of faces (part of a solid)."""
    def __init__(self, shell):
        self.faces = np.array(shell['faces'], dtype=object)
        self.orientation_wrt_solid = shell['orientation_wrt_solid']
        self.solids = None

        self.id = shell['id']




class Solid(Topology):
    """Solid: top-level collection of shells comprising a part's volume."""
    def __init__(self, solid):
        self.shells = np.array(solid['shells'])
        self.id = solid['id']



"""
Computations for winding number and related operations for curves/surfaces.
"""

import numpy as np

def winding_number(curve_uv_values, surface_uv_values):
    """ Compute the winding number for a polyline and surface UV values efficiently. """
    a_values = curve_uv_values[:-1]
    b_values = curve_uv_values[1:]
    a = a_values[:, np.newaxis] - surface_uv_values
    b = b_values[:, np.newaxis] - surface_uv_values

    det = a[..., 0] * b[..., 1] - a[..., 1] * b[..., 0]

    dot = np.einsum('ijk,ijk->ij', a, b)
    winding_number_result = np.sum(np.arctan2(det, dot), axis=0) / (2 * np.pi)
    return winding_number_result.reshape(-1, 1)

def find_surface_uv_for_curve(surface_points, surface_uv_values, curve_points):
        """
        Calculate the nearest UV values on a surface for a given set of curve points.

        Args:
        surface_points (np.ndarray): Points on the surface.
        surface_uv_values (np.ndarray): UV values on the surface.
        curve_points (np.ndarray): Points on the curve.

        Returns:
        np.ndarray: UV values on the surface closest to the curve points.
        """
        # Calculate the nearest surface point for each curve point
        # nearest_3d_surface_points, curve_indexes = self._calculate_nearest_surface_points(surface_points, curve_points)
        from scipy.spatial import KDTree
        tree = KDTree(surface_points)
        _, curve_indexes = tree.query(curve_points)

        if type(curve_indexes) == np.int64:
            curve_indexes = [curve_indexes]

        if surface_uv_values.size > 0 and max(curve_indexes) < len(surface_uv_values):
            surface_uv_near_curve = surface_uv_values[curve_indexes]
        else:
            # Handle the case where surface_uv_values is empty or too small
            # This could be setting surface_uv_near_curve to an empty array
            # or some other default value, depending on your application's needs
            surface_uv_near_curve = np.array([])

        return surface_uv_near_curve



"""Command-line interface for converting ABS geometry to PLY or Pickle formats."""
from __future__ import annotations

import os
import numpy as np
import argparse
from concurrent.futures import ProcessPoolExecutor, as_completed
from tqdm import tqdm
from joblib import dump
import traceback
from abs.utils import read_parts
from abs.part_processor import sample_parts

def get_normal_func(part, topo, points):
    """Return normal vectors for faces, None for other topology elements."""
    if topo.is_face():
        return topo.normal(points)
    else:
        return None

def to_ply_main():
    """Console script entry point for converting to PLY."""
    parser = argparse.ArgumentParser(prog="abs-to-ply",
                                     description="Sample points and normals from ABS geometry and output as PLY files.")
    parser.add_argument("input_path", help="Path to an input .hdf5 file or a directory containing .hdf5 files")
    parser.add_argument("output_dir", help="Directory to save output PLY files")
    parser.add_argument("-n", "--num-samples", dest="num_samples", type=int, default=2000,
                        help="Number of sample points per part (default: 2000)")
    parser.add_argument("-j", "--jobs", dest="max_workers", type=int, default=4,
                        help="Number of parallel worker processes (default: 4)")
    args = parser.parse_args()
    input_path = args.input_path
    output_dir = args.output_dir
    num_samples = args.num_samples
    max_workers = args.max_workers
    os.makedirs(output_dir, exist_ok=True)
    # Determine if input is a single file or directory of files
    files_to_process = []
    if os.path.isdir(input_path):
        # Process all .hdf5 files in directory
        files_to_process = [os.path.join(input_path, f) for f in os.listdir(input_path) if f.endswith('.hdf5')]
    elif os.path.isfile(input_path):
        files_to_process = [input_path]
    else:
        parser.error(f"Input path {input_path} does not exist.")
    error_files = []
    # Define worker function
    def process_and_save_ply(file_path):
        try:
            parts = read_parts(file_path)
            P, S = sample_parts(parts, num_samples, get_normal_func)
            base_name = os.path.basename(file_path)
            for i in range(len(P)):
                points = P[i]
                normals = S[i]
                if len(points) != len(normals):
                    raise RuntimeError(f"Points/normals length mismatch for part {i} in {file_path}")
                out_file = os.path.join(output_dir, f"{base_name}_part_{i}.ply")
                with open(out_file, 'w') as f:
                    # Write PLY header
                    f.write("ply\n")
                    f.write("format ascii 1.0\n")
                    f.write(f"element vertex {len(points)}\n")
                    f.write("property float x\n")
                    f.write("property float y\n")
                    f.write("property float z\n")
                    f.write("property float nx\n")
                    f.write("property float ny\n")
                    f.write("property float nz\n")
                    f.write("end_header\n")
                    # Write each vertex and normal
                    for p, n in zip(points, normals):
                        f.write(f"{p[0]} {p[1]} {p[2]} {n[0]} {n[1]} {n[2]}\n")
        except Exception as e:
            # Print traceback for debugging
            traceback.print_exc()
            raise
    # Process files in parallel
    if max_workers > 1 and len(files_to_process) > 1:
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            futures = {executor.submit(process_and_save_ply, fp): fp for fp in files_to_process}
            for future in tqdm(as_completed(futures), total=len(futures), desc="Processing files"):
                file_path = futures[future]
                try:
                    future.result()
                except Exception as exc:
                    error_files.append(os.path.basename(file_path))
                    print(f"Error processing file {file_path}: {exc}")
    else:
        # Single-threaded processing
        for file_path in tqdm(files_to_process, desc="Processing files"):
            try:
                process_and_save_ply(file_path)
            except Exception as exc:
                error_files.append(os.path.basename(file_path))
                print(f"Error processing file {file_path}: {exc}")
    if error_files:
        print("\nThe following files failed to process:")
        for fname in error_files:
            print(fname)
    else:
        print("\nAll files processed successfully.")

def to_pickle_main():
    """Console script entry point for converting to Pickle."""
    parser = argparse.ArgumentParser(prog="abs-to-pickle",
                                     description="Sample points and normals from ABS geometry and output as Pickle files.")
    parser.add_argument("input_path", help="Path to an input .hdf5 file or a directory containing .hdf5 files")
    parser.add_argument("output_dir", help="Directory to save output .pkl files")
    parser.add_argument("-n", "--num-samples", dest="num_samples", type=int, default=2000,
                        help="Number of sample points per part (default: 2000)")
    parser.add_argument("-j", "--jobs", dest="max_workers", type=int, default=4,
                        help="Number of parallel worker processes (default: 4)")
    args = parser.parse_args()
    input_path = args.input_path
    output_dir = args.output_dir
    num_samples = args.num_samples
    max_workers = args.max_workers
    os.makedirs(output_dir, exist_ok=True)
    files_to_process = []
    if os.path.isdir(input_path):
        files_to_process = [os.path.join(input_path, f) for f in os.listdir(input_path) if f.endswith('.hdf5')]
    elif os.path.isfile(input_path):
        files_to_process = [input_path]
    else:
        parser.error(f"Input path {input_path} does not exist.")
    error_files = []
    def process_and_save_pickle(file_path):
        try:
            parts = read_parts(file_path)
            P, S = sample_parts(parts, num_samples, get_normal_func)
            base_name = os.path.basename(file_path)
            for i in range(len(P)):
                data = {
                    'file': base_name,
                    'part': i,
                    'points': P[i],
                    'normals': S[i]
                }
                out_file = os.path.join(output_dir, f"{base_name}_part_{i}.pkl")
                dump(data, out_file)
        except Exception as e:
            traceback.print_exc()
            raise
    if max_workers > 1 and len(files_to_process) > 1:
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            futures = {executor.submit(process_and_save_pickle, fp): fp for fp in files_to_process}
            for future in tqdm(as_completed(futures), total=len(futures), desc="Processing files"):
                file_path = futures[future]
                try:
                    future.result()
                except Exception as exc:
                    error_files.append(os.path.basename(file_path))
                    print(f"Error processing file {file_path}: {exc}")
    else:
        for file_path in tqdm(files_to_process, desc="Processing files"):
            try:
                process_and_save_pickle(file_path)
            except Exception as exc:
                error_files.append(os.path.basename(file_path))
                print(f"Error processing file {file_path}: {exc}")
    if error_files:
        print("\nThe following files failed to process:")
        for fname in error_files:
            print(fname)
    else:
        print("\nAll files processed successfully.")



"""
Functions for processing Shape parts: sampling points and computing normals.
"""

import numpy as np
from . import sampler
from abspy import poisson_disk_downsample



def estimate_total_surface_area(part):
    """Estimate the total surface area of all faces in a Shape part."""
    total_area = 0.0
    for face in part.Solid.faces:
        total_area += face.get_area()
    return total_area

def estimate_total_curve_length(part):
    """Estimate the total curve length of all edges in a Shape part."""
    total_length = 0.0
    for edge in part.Solid.edges:
        total_length += edge.get_length()
    return total_length


def process_part(part, num_samples, lambda_func, points_ratio=5):
    """
    Sample points on all surfaces and curves of a shape part according to specified number.
    Uses an iterative strategy with oversampling (points_ratio) and Poisson disk downsampling to refine points.
    The lambda_func is a function(part, topo_entity, param_points) that returns associated values (e.g., normals).
    """
    initial_num_points = points_ratio * num_samples
    num_points = initial_num_points
    total_area = estimate_total_surface_area(part)
    total_length = estimate_total_curve_length(part)
    pts = np.zeros((0, 3))
    ss = []
    while True:
        current_pts = []
        current_ss = []
        # Sample points on each face (surface)
        for face in part.Solid.faces:
            n_surf = int(np.ceil((face.get_area() / total_area) * num_points))
            uv_points, pt = sampler.random_sample(face, n_surf, min_pts=2)
            s = lambda_func(part, face, uv_points)
            if s is not None:
                # Transform points to surface's local coordinate system
                R = np.linalg.inv(face.surface.transform[:, :3])
                t = face.surface.transform[:, 3]
                pt = (pt - t) @ R.T
                # Filter out points outside trimming loops
                index = face.filter_outside_points(uv_points)
                current_pts.append(pt[index, :])
                if isinstance(s, list):
                    # If lambda_func returns multiple arrays
                    for i, val in enumerate(s):
                        if len(val) != len(pt):
                            val = np.full((pt.shape[0], pt.shape[1]), val)
                        elif val.shape[1] != pt.shape[1]:
                            val = np.tile(val, (1, pt.shape[1]))
                        if len(current_ss) == 0:
                            current_ss = [[] for _ in range(len(s))]
                        current_ss[i].extend(val[index, :])
                else:
                    if len(s) != len(pt):
                        s = np.full((pt.shape[0], pt.shape[1]), s)
                    elif s.shape[1] != pt.shape[1]:
                        s = np.tile(s, (1, pt.shape[1]))
                    current_ss.append(s[index, :])
        # Sample points on each edge (curve)
        for edge in part.Solid.edges:
            if edge.curve3d is None:
                continue
            n_edge = int(np.ceil((edge.get_length() / total_length) * num_points))
            uv_points, pt = sampler.random_sample(edge, n_edge, min_pts=2)
            s = lambda_func(part, edge, uv_points)
            if s is not None:
                R = np.linalg.inv(edge.curve3d.transform[:, :3])
                t = edge.curve3d.transform[:, 3]
                pt = (pt - t) @ R.T
                # Edges have no trimming curves; include all sampled points
                index = np.ones(uv_points.shape[0], dtype=bool)
                current_pts.append(pt[index, :])
                if isinstance(s, list):
                    for i, val in enumerate(s):
                        if len(val) != len(pt):
                            val = np.full((pt.shape[0], pt.shape[1]), val)
                        elif val.shape[1] != pt.shape[1]:
                            val = np.tile(val, (1, pt.shape[1]))
                        if len(current_ss) == 0:
                            current_ss = [[] for _ in range(len(s))]
                        current_ss[i].extend(val[index, :])
                else:
                    if len(s) != len(pt):
                        s = np.full((pt.shape[0], pt.shape[1]), s)
                    elif s.shape[1] != pt.shape[1]:
                        s = np.tile(s, (1, pt.shape[1]))
                    current_ss.append(s[index, :])
        # Combine all collected points and associated data
        if len(current_pts) == 0:
            pts = np.zeros((0, 3))
            ss = []
        else:
            pts = np.concatenate(current_pts, axis=0)
            if isinstance(current_ss, list) and len(current_ss) > 0 and isinstance(current_ss[0], list):
                ss = current_ss.copy()
            else:
                ss = np.concatenate(current_ss, axis=0) if len(current_ss) > 0 else []
        # Stop when enough points collected or nothing collected
        if pts.shape[0] >= initial_num_points or pts.shape[0] == 0:
            break
        else:
            # Increase num_points for next iteration (upscale by factor to approach initial count)
            num_points = np.ceil(num_points * initial_num_points / max(len(pts), 1) * 1.2)
    # If no points at all, return empty structures
    if pts.shape[0] == 0:
        return pts, pts
    # Poisson disk downsample to exactly num_samples points
    indices = poisson_disk_downsample(pts, num_samples)
    if len(indices) < num_samples:
        remaining_idx = [i for i in range(len(pts)) if i not in indices]
        additional_indices = np.random.choice(remaining_idx, num_samples - len(indices), replace=False)
        indices = np.concatenate([indices, additional_indices])
    elif len(indices) > num_samples:
        indices = np.random.choice(indices, num_samples, replace=False)
    if isinstance(ss, list):
        # Multiple associated arrays (e.g., normals in separate array)
        new_ss = [[sublist[i] for i in indices] for sublist in ss]
        return pts[indices], new_ss
    else:
        return pts[indices], ss[indices] if isinstance(ss, np.ndarray) else ss


def sample_parts(parts, num_samples, lambda_func, points_ratio=5):
    """Process a list of parts by sampling each part and returning lists of points and values."""
    pts_list = []
    ss_list = []
    for part in parts:
        pts, ss = process_part(part, num_samples, lambda_func, points_ratio)
        pts_list.append(np.array(pts))
        if isinstance(ss, list):
            ss_list.extend([np.array(sublist) for sublist in ss])
        else:
            ss_list.append(np.array(ss))
    return pts_list, ss_list




"""
ABS Geometry Processing Package
"""

"""
ABS geometry-processing toolkit.

Public API
----------
read_parts, read_meshes, sample_parts, poisson_disk_downsample
"""

from importlib import import_module
from warnings import warn
from importlib.metadata import version, PackageNotFoundError

# ── public sub-modules (relative imports!) ────────────────────
from abs import utils, sampler, part_processor
from abs.utils import read_parts, read_meshes
from abs.part_processor import sample_parts

# ── optional C++ extension (abspy) ────────────────────────────
try:
    _abspy = import_module("abspy")
    poisson_disk_downsample = _abspy.poisson_disk_downsample
    __version__ = version("abs-hdf5")
except ModuleNotFoundError as exc:
    poisson_disk_downsample = None
    warn(
        "C++ extension 'abspy' not found; blue-noise down-sampling disabled. "
        f"(Original error: {exc})",
        RuntimeWarning,
        stacklevel=2,
    )
except PackageNotFoundError:
    __version__ = "0.1.1"

__all__ = [
    "read_parts",
    "read_meshes",
    "sample_parts",
    "poisson_disk_downsample",
]



"""Unit test package for abs."""
import importlib, sys as _sys
for _name in ("test_utilities", "test_geometry"):
    try:
        _mod = importlib.import_module(f".{_name}", package=__name__)
        _sys.modules[_name] = _mod          # alias at top-level
    except ModuleNotFoundError:
        pass



"""
Utility functions for reading shape parts from files and saving data in common formats.
"""
from pathlib import Path
import meshio as mio
import numpy as np
import h5py
from .shape import Shape


def read_parts(file_path):
    """Read all parts from an HDF5 file and construct Shape objects for each."""
    f = h5py.File(file_path, 'r')
    part = f['parts'].values()

    parts = []
    for i, p in enumerate(part):
        s = Shape(p['geometry'], p['topology'])
        parts.append(s)

    return parts


def read_meshes(file_path):
    """Read pre-computed mesh data (points and triangles) for each face of each part from an HDF5 file."""
    f = h5py.File(file_path, 'r')
    part = f['parts'].values()

    meshes = []
    for i, p in enumerate(part):
        s = Shape(p['geometry'], p['topology'])
        # If shape has no faces, append empty
        if not hasattr(s, 'Solid') or not hasattr(s.Solid, 'faces') or not s.Solid.faces:
            meshes.append([])
            continue
        mesh_group = p['mesh']
        current_mesh = [None] * len(s.Solid.faces)
        for key in mesh_group:
            submesh = mesh_group[key]
            vertices = submesh['points']
            faces = submesh['triangle']

            current_mesh[int(key)] = {
                'points': vertices,
                'triangle': faces
            }
        meshes.append(current_mesh)

    return meshes


# Functions for saving data (if needed)
def save_obj(filename, pts):
    '''
    Save a set of 3D points to an .obj file.
    '''
    with open(filename, "w") as f:
        if pts.shape[1] == 2:
            [f.write(f"v {pts[i, 0]} {pts[i, 1]} 0\n") for i in range(pts.shape[0])]
        else:
            [f.write(f"v {pts[i, 0]} {pts[i, 1]} {pts[i, 2]}\n") for i in range(pts.shape[0])]


def save_obj_mesh(filename, pts, faces):
    '''
    Save a set of 3D points and faces to an .obj file.
    '''
    if pts.shape[0] == 0:
        print("Skipping saving meshes: mesh is empty")
        return

    with open(filename, "w") as f:
        if pts.shape[1] == 2:
            [f.write(f"v {pts[i, 0]} {pts[i, 1]} 0\n") for i in range(pts.shape[0])]
        else:
            [f.write(f"v {pts[i, 0]} {pts[i, 1]} {pts[i, 2]}\n") for i in range(pts.shape[0])]

        [f.write(f"f {faces[i, 0]+1} {faces[i, 1]+1} {faces[i, 2]+1}\n") for i in range(faces.shape[0])]


def save_ply(filename, P, normals=None):
    '''
    Save a set of 3D points to a .ply file. Optionally, also save normals.
    '''
    total_points = []
    total_normals = []

    # for idx, part in enumerate(P):
    #     if normals:
    #         normal = normals[idx]
    #     for i, pts in enumerate(part):
    #         if normals:
    #             if pts.shape[0] != normal.shape[0]:
    #                 raise ValueError("The number of points and normals must be the same")
    #             if pts.shape[1] != 3 or normal.shape[1] != 3:
    #                 raise ValueError("Both pts and normals must have shape (n, 3)")
    #             total_points.append(pts)
    #             total_normals.append(normal)
    #         else:
    #             if pts.shape[1] != 3:
    #                 raise ValueError("Points must have shape (n, 3)")
    #             total_points.append(pts)


    for i, pts in enumerate(P):
        if (pts.shape[0] == 0):
            continue
        if normals:
            normal = normals[i]
            if pts.shape[0] != normal.shape[0]:
                raise ValueError("The number of points and normals must be the same")
            if pts.shape[1] != 3 or normal.shape[1] != 3:
                raise ValueError("Both pts and normals must have shape (n, 3)")
            total_points.append(pts)
            total_normals.append(normal)
        else:
            if pts.shape[1] != 3:
                raise ValueError("Points must have shape (n, 3)")
            total_points.append(pts)

    new_pts = np.asarray(total_points)
    if new_pts.shape[2] == 1:
        new_pts = np.squeeze(new_pts, axis=0)
    else:
        new_pts = np.vstack(new_pts)

    if total_normals:
        new_normal = np.asarray(total_normals)
        if new_normal.shape[2] == 1:
            new_normal = np.squeeze(new_normal, axis=0)
        else:
            new_normal = np.vstack(new_normal)

        if new_pts.shape[0] != new_normal.shape[0]:
            raise ValueError("The number of points and normals must be the same")

        if new_pts.shape[1] != 3 or new_normal.shape[1] != 3:
            raise ValueError("Both pts and normals must have shape (n, 3)")

        data = np.hstack((new_pts, new_normal))

        header = f"""ply
format ascii 1.0
element vertex {data.shape[0]}
property float x
property float y
property float z
property float nx
property float ny
property float nz
end_header
"""
    else:
        if new_pts.shape[1] != 3:
            raise ValueError("Points must have shape (n, 3)")

        data = new_pts

        header = f"""ply
format ascii 1.0
element vertex {data.shape[0]}
property float x
property float y
property float z
end_header
"""
    with open(filename, 'w') as f:
        f.write(header)
        if normals is not None:
            np.savetxt(f, data, fmt='%f %f %f %f %f %f')
        else:
            np.savetxt(f, data, fmt='%f %f %f')

def save_to_xyz(points, filename):
    """
    Save 3D points to an .xyz file.
    """
    with open(filename, 'w') as f:
        for point in points:
            # Write each point as X Y Z in a new line
            f.write(f"{point[0]} {point[1]} {point[2]}\n")


def save_vtu(save_file_path , P):
    m = mio.Mesh(P, cells={"triangle":np.array([np.arange(P.shape[0]), np.arange(P.shape[0]), np.arange(P.shape[0])]).T})
    m.write(save_file_path)




def get_mesh(meshes):
    global_vertices = []
    global_faces = []
    vertex_offset = 0

    for mesh in meshes:
        for sub_mesh in mesh:
            if sub_mesh is None:
                continue

            vertices = sub_mesh["points"][:]
            if len(vertices) == 0:
                continue

            global_vertices.append(vertices)
            faces = sub_mesh["triangle"][:] + vertex_offset
            global_faces.append(faces)

            vertex_offset += vertices.shape[0]

    if global_vertices:
        global_vertices = np.vstack(global_vertices)
    else:
        global_vertices = np.empty((0, 3))

    if global_faces:
        global_faces = np.vstack(global_faces)
    else:
        global_faces = np.empty((0, 3), dtype=int)

    return global_vertices, global_faces



def get_mesh_part(mesh):
    global_vertices = []
    global_faces = []
    vertex_offset = 0

    for key in mesh:

        sub_mesh = mesh[key]

        vertices = sub_mesh["points"][:]
        if len(vertices) == 0:
            continue
        global_vertices.append(vertices)

        faces = sub_mesh["triangle"][:] + vertex_offset
        global_faces.append(faces)

        vertex_offset += vertices.shape[0]

    global_vertices = np.vstack(global_vertices)
    global_faces = np.vstack(global_faces)

    return global_vertices, global_faces





"""
Surface classes for various surface types (Plane, Cylinder, Cone, Sphere, Torus, etc.).
Provide methods to sample points on surfaces and compute derivatives and normals.
"""

import numpy as np
from geomdl import BSpline, NURBS
from geomdl import operations
from .curve import create_curve
from scipy.stats import skew


def create_surface(surface_data, compute_index=True):
    if compute_index:
        index = int(surface_data.name.split("/")[-1])
    else:
        index = None

    surface_type = surface_data.get('type')[()].decode('utf-8')
    surface_map = {
        'Plane': Plane,
        'Cylinder': Cylinder,
        'Cone': Cone,
        'Sphere': Sphere,
        'Torus': Torus,
        'BSpline': BSplineSurface,
        'Extrusion': Extrusion,
        'Revolution': Revolution,
        'Offset': Offset
    }
    surface_class = surface_map.get(surface_type)
    if surface_class:
        return index, surface_class(surface_data)
    else:
        # Unsupported surface type
        return index, None

class Surface:
    """Base class for all surface types."""
    def sample(self, points):
        raise NotImplementedError("Sample method must be implemented by subclasses")
    def derivative(self, points, order=1):
        raise NotImplementedError("Derivative method must be implemented by subclasses")
    def normal(self, points):
        derivatives = self.derivative(points, order=1)
        normals = np.cross(derivatives[:, :, 0], derivatives[:, :, 1])
        normals = normals / np.linalg.norm(normals, axis=1)[:, np.newaxis]
        return normals

    def get_area(self):
        if getattr(self, "area", None) is not None and self.area != -1:
            return self.area
        # Approximate area via 4x4 Gauss-Legendre quadrature

        x, w = np.polynomial.legendre.leggauss(4)
        pts = np.array(np.meshgrid(x, x, indexing='ij')).reshape(2, -1).T+1
        pts *= 0.5 * (self.trim_domain[:, 1] - self.trim_domain[:, 0])
        pts += self.trim_domain[:, 0]
        weights = (w * w[:, None]).ravel()

        dd = self.derivative(pts)
        EE = np.sum(dd[:, :, 0] * dd[:, :, 0], axis=1)
        FF = np.sum(dd[:, :, 0] * dd[:, :, 1], axis=1)
        GG = np.sum(dd[:, :, 1] * dd[:, :, 1], axis=1)

        self.area = np.sum(np.sqrt(EE * GG - FF ** 2)*weights)*np.prod(self.trim_domain[:, 1] - self.trim_domain[:, 0]) / 4

        return self.area


class Plane(Surface):
    """Plane surface."""
    def __init__(self, plane):
        self.location = np.array(plane.get('location')[()]).reshape(-1, 1).T
        self.coefficients = np.array(plane.get('coefficients')[()]).reshape(-1, 1).T
        self.trim_domain = np.array(plane.get('trim_domain')[()])
        self.x_axis = np.array(plane.get('x_axis')[()]).reshape(-1, 1).T
        self.y_axis = np.array(plane.get('y_axis')[()]).reshape(-1, 1).T
        self.z_axis = np.array(plane.get('z_axis')[()]).reshape(-1, 1).T
        self.transform = np.array(plane.get('transform')[()])
        self.area = -1
        self.shape_name = plane.get('type')[()].decode('utf8')

    def sample(self, sample_points):
        if sample_points.size == 0:
            return self.location
        plane_points = self.location.T + sample_points[:, 0] * self.x_axis.T + sample_points[:, 1] * self.y_axis.T
        return plane_points.T

    def derivative(self, sample_points, order=1):
        if order == 0:
            return self.sample(sample_points)
        elif order == 1:
            # Partial derivatives in u and v directions
            deriv = np.zeros((sample_points.shape[0], 3, 2))
            deriv[:, :, 0] = self.x_axis
            deriv[:, :, 1] = self.y_axis
            return deriv
        elif order == 2:
            # Second derivative is zero
            return np.zeros((sample_points.shape[0], 3, 2, 2))
        else:
            raise ValueError("Order must be 0, 1, or 2")

    def normal(self, sample_points):
        normal_vector = np.cross(self.x_axis.squeeze(), self.y_axis.squeeze())
        normal_vector_normalized = normal_vector / np.linalg.norm(normal_vector)
        normals = np.tile(normal_vector_normalized, (sample_points.shape[0], 1))
        return normals


class Cylinder(Surface):
    """Cylindrical surface."""
    def __init__(self, cylinder):

        self.location = np.array(cylinder.get('location')[()]).reshape(-1, 1).T
        self.radius = float(cylinder.get('radius')[()])
        self.coefficients = np.array(cylinder.get('coefficients')[()]).reshape(-1, 1).T
        self.trim_domain = np.array(cylinder.get('trim_domain')[()])
        self.x_axis = np.array(cylinder.get('x_axis')[()]).reshape(-1, 1).T
        self.y_axis = np.array(cylinder.get('y_axis')[()]).reshape(-1, 1).T
        self.z_axis = np.array(cylinder.get('z_axis')[()]).reshape(-1, 1).T
        self.transform = np.array(cylinder.get('transform')[()])
        self.area = -1
        self.shape_name = cylinder.get('type')[()].decode('utf8')

    def sample(self, sample_points):

        if sample_points.size == 0:
            return self.location

        cylinder_points = self.location.T + self.radius * np.cos(sample_points[:, 0]) * self.x_axis.T + \
                          self.radius * np.sin(sample_points[:, 0]) * self.y_axis.T + sample_points[:, 1] * \
                          self.z_axis.T
        return cylinder_points.T

    def derivative(self, sample_points, order=1):
        if order == 0:
            return self.sample(sample_points)
        elif order == 1:
            dev = np.zeros((sample_points.shape[0], 3, 2))
            dev[:, :, 0] = self.radius * \
                           (-np.sin(sample_points[:, 0]) * self.x_axis.T +
                            np.cos(sample_points[:, 0]) * self.y_axis.T).T
            dev[:, :, 1] = self.z_axis
            return dev
        elif order == 2:
            dev = np.zeros((sample_points.shape[0], 3, 2, 2))
            dev[:, :, 0, 0] = (-self.radius * (np.cos(sample_points[:, 0]) * self.x_axis.T +
                                                np.sin(sample_points[:, 0]) * self.y_axis.T)).T
            return dev
        else:
            raise ValueError("Order must be 0, 1, or 2")


class Cone(Surface):
    """Conical surface."""
    def __init__(self, cone):
        self.location = np.array(cone.get('location')[()]).reshape(-1, 1).T
        self.radius = float(cone.get('radius')[()])
        self.coefficients = np.array(cone.get('coefficients')[()]).reshape(-1, 1).T
        self.trim_domain = np.array(cone.get('trim_domain')[()])
        self.apex = np.array(cone.get('apex')[()]).reshape(-1, 1).T
        self.angle = float(cone.get('angle')[()])
        self.x_axis = np.array(cone.get('x_axis')[()]).reshape(-1, 1).T
        self.y_axis = np.array(cone.get('y_axis')[()]).reshape(-1, 1).T
        self.z_axis = np.array(cone.get('z_axis')[()]).reshape(-1, 1).T
        self.transform = np.array(cone.get('transform')[()])
        self.area = -1
        self.shape_name = cone.get('type')[()].decode('utf8')

    def sample(self, sample_points):

        if sample_points.size == 0:
            return self.apex
        cone_points = self.location.T + (self.radius + sample_points[:, 1] * np.sin(self.angle)) * \
                      (np.cos(sample_points[:, 0]) * self.x_axis.T + np.sin(sample_points[:, 0]) * self.y_axis.T) \
                      + sample_points[:, 1] * np.cos(self.angle) * self.z_axis.T
        return cone_points.T

    def derivative(self, sample_points, order=1):
        if order == 0:
            return self.sample(sample_points)
        elif order == 1:
            dev = np.zeros((sample_points.shape[0], 3, 2))
            dev[:, :, 0] = ((self.radius + sample_points[:, 1] * np.sin(self.angle)) *
                            (-np.sin(sample_points[:, 0]) * self.x_axis.T +
                             np.cos(sample_points[:, 0]) * self.y_axis.T)).T
            dev[:, :, 1] = (np.sin(self.angle) *
                            (np.cos(sample_points[:, 0]) * self.x_axis.T + np.sin(sample_points[:, 0]) *
                             self.y_axis.T) + np.cos(self.angle) * self.z_axis.T).T
            return dev
        elif order == 2:
            dev = np.zeros((sample_points.shape[0], 3, 2, 2))
            dev[:, :, 0, 0] = ((self.radius + sample_points[:, 1] * np.sin(self.angle)) *
                               (-np.cos(sample_points[:, 0]) * self.x_axis.T -
                                np.sin(sample_points[:, 0]) * self.y_axis.T)).T
            dev[:, :, 0, 1] = (np.sin(self.angle) * (-np.sin(sample_points[:, 0]) *
                                                      self.x_axis.T + np.cos(sample_points[:, 0]) * self.y_axis.T).T)
            dev[:, :, 1, 0] = (np.sin(self.angle) * (-np.sin(sample_points[:, 0]) *
                                                      self.x_axis.T + np.cos(sample_points[:, 0]) * self.y_axis.T).T)
            return dev
        else:
            raise ValueError("Order must be 0, 1, or 2")


class Sphere(Surface):
    """Spherical surface."""
    def __init__(self, sphere):

        self.location = np.array(sphere.get('location')[()]).reshape(-1, 1).T
        self.radius = float(sphere.get('radius')[()])
        self.coefficients = np.array(sphere.get('coefficients')[()]).reshape(-1, 1).T
        self.trim_domain = np.array(sphere.get('trim_domain')[()])
        self.x_axis = np.array(sphere.get('x_axis')[()]).reshape(-1, 1).T
        self.y_axis = np.array(sphere.get('y_axis')[()]).reshape(-1, 1).T
        if 'z_axis' in sphere:
            self.z_axis = np.array(sphere.get('z_axis')[()]).reshape(-1, 1).T
        else:
            self.z_axis = np.cross(self.x_axis, self.y_axis)
        self.transform = np.array(sphere.get('transform')[()])
        self.area = -1
        self.shape_name = sphere.get('type')[()].decode('utf8')

    def sample(self, sample_points):
        if sample_points.size == 0:
            return self.location
        sphere_points = self.location.T + self.radius * np.cos(sample_points[:, 1]) * \
                        (np.cos(sample_points[:, 0]) * self.x_axis.T + np.sin(sample_points[:, 0]) * self.y_axis.T) \
                        + self.radius * np.sin(sample_points[:, 1]) * self.z_axis.T
        return sphere_points.T

    def derivative(self, sample_points, order=1):
        if order == 0:
            return self.sample(sample_points)
        elif order == 1:
            dev = np.zeros((sample_points.shape[0], 3, 2))
            dev[:, :, 0] = (self.radius * np.cos(sample_points[:, 1]))[:, np.newaxis] * \
                           (-np.sin(sample_points[:, 0]) * self.x_axis.T +
                            np.cos(sample_points[:, 0]) * self.y_axis.T).T
            dev[:, :, 1] = -self.radius * np.sin(sample_points[:, 1])[:, np.newaxis] * \
                           (np.cos(sample_points[:, 0]) * self.x_axis.T +
                            np.sin(sample_points[:, 0]) * self.y_axis.T).T + \
                           (self.radius * np.cos(sample_points[:, 1]) * self.z_axis.T).T
            return dev
        elif order == 2:
            dev = np.zeros((sample_points.shape[0], 3, 2, 2))
            dev[:, :, 0, 0] = (self.radius * np.cos(sample_points[:, 1]) *
                               (-np.cos(sample_points[:, 0]) *
                                self.x_axis.T - np.sin(sample_points[:, 0]) * self.y_axis.T)).T
            dev[:, :, 0, 1] = (-self.radius * np.sin(sample_points[:, 1])
                               * (-np.sin(sample_points[:, 0]) *
                                  self.x_axis.T + np.cos(sample_points[:, 0]) * self.y_axis.T)).T
            dev[:, :, 1, 0] = (-self.radius * np.sin(sample_points[:, 1]) *
                               (-np.sin(sample_points[:, 0]) * self.x_axis.T +
                                np.cos(sample_points[:, 0]) * self.y_axis.T)).T
            dev[:, :, 1, 1] = (-self.radius * np.cos(sample_points[:, 1]) *
                               (np.cos(sample_points[:, 0]) * self.x_axis.T +
                                np.sin(sample_points[:, 0]) * self.y_axis.T) -
                               self.radius * np.sin(sample_points[:, 1]) * self.z_axis.T).T
            return dev
        else:
            raise ValueError("Order must be 0, 1, or 2")

    def normal(self, sample_points):
        if sample_points.size == 0:
            return np.array([])
        sphere_points = self.sample(sample_points)
        normals = sphere_points - self.location
        normals = normals / np.linalg.norm(normals, axis=1)[:, np.newaxis]
        return normals


class Torus(Surface):
    """Torus (donut-shaped) surface."""
    def __init__(self, torus):

        self.location = np.array(torus.get('location')[()]).reshape(-1, 1).T
        self.max_radius = float(torus.get('max_radius')[()])
        self.min_radius = float(torus.get('min_radius')[()])
        self.trim_domain = np.array(torus.get('trim_domain')[()])
        self.x_axis = np.array(torus.get('x_axis')[()]).reshape(-1, 1).T
        self.y_axis = np.array(torus.get('y_axis')[()]).reshape(-1, 1).T
        self.z_axis = np.array(torus.get('z_axis')[()]).reshape(-1, 1).T
        self.transform = np.array(torus.get('transform')[()])
        self.area = -1
        self.shape_name = torus.get('type')[()].decode('utf8')

    def sample(self, sample_points):
        if sample_points.size == 0:
            return self.location
        torus_points = self.location.T + (self.max_radius + self.min_radius * np.cos(sample_points[:, 1])) * \
                       (np.cos(sample_points[:, 0]) * self.x_axis.T + np.sin(sample_points[:, 0]) * self.y_axis.T) \
                       + self.min_radius * np.sin(sample_points[:, 1]) * self.z_axis.T
        return torus_points.T

    def derivative(self, sample_points, order=1):
        if order == 0:
            return self.sample(sample_points)
        elif order == 1:
            dev = np.zeros((sample_points.shape[0], 3, 2))

            dev[:, :, 0] = ((self.max_radius + self.min_radius * np.cos(sample_points[:, 1])) *
                            (-np.sin(sample_points[:, 0]) * self.x_axis.T +
                             np.cos(sample_points[:, 0]) * self.y_axis.T)).T

            dev[:, :, 1] = (((-self.min_radius * np.sin(sample_points[:, 1])) *
                             (np.cos(sample_points[:, 0]) * self.x_axis.T +
                              np.sin(sample_points[:, 0]) * self.y_axis.T)) +
                            (self.min_radius * np.cos(sample_points[:, 1])) * self.z_axis.T).T
            return dev
        elif order == 2:
            dev = np.zeros((sample_points.shape[0], 3, 2, 2))
            dev[:, :, 0, 0] = ((self.max_radius + self.min_radius * np.cos(sample_points[:, 1])) *
                               (-np.cos(sample_points[:, 0]) * self.x_axis.T -
                                np.sin(sample_points[:, 0]) * self.y_axis.T)).T

            dev[:, :, 0, 1] = (-self.min_radius * np.sin(sample_points[:, 1]) *
                               (-np.sin(sample_points[:, 0]) * self.x_axis.T +
                                np.cos(sample_points[:, 0]) * self.y_axis.T)).T

            dev[:, :, 1, 0] = (-self.min_radius * np.sin(sample_points[:, 1]) *
                               (-np.sin(sample_points[:, 0]) * self.x_axis.T +
                                np.cos(sample_points[:, 0]) * self.y_axis.T)).T

            dev[:, :, 1, 1] = ((-self.min_radius * np.cos(sample_points[:, 1]) *
                                (np.cos(sample_points[:, 0]) * self.x_axis.T + np.sin(sample_points[:, 0])
                                 * self.y_axis.T)) - (self.min_radius *
                                                       np.sin(sample_points[:, 1]) * self.z_axis.T)).T
            return dev
        else:
            raise ValueError("Order must be 0, 1, or 2")


class BSplineSurface(Surface):
    """B-spline or NURBS surface (possibly trimmed)."""
    def __init__(self, bspline_surface):

        self.continuity = int(bspline_surface.get('continuity')[()])
        self.face_domain = np.array(bspline_surface.get('face_domain')[()]).reshape(-1, 1).T
        self.is_trimmed = bool(bspline_surface.get('is_trimmed')[()])
        self.poles = np.array(bspline_surface.get('poles')[()])
        self.trim_domain = np.array(bspline_surface.get('trim_domain')[()])
        if len(self.trim_domain.shape) == 1:
            self.trim_domain = np.reshape(self.trim_domain, [2, 2])
        self.u_closed = bool(bspline_surface.get('u_closed')[()])
        self.u_degree = int(bspline_surface.get('u_degree')[()])
        self.u_knots = np.array(bspline_surface.get('u_knots')[()]).reshape(-1, 1).T
        self.u_rational = bool(bspline_surface.get('u_rational')[()])
        self.v_closed = bool(bspline_surface.get('v_closed')[()])
        self.v_degree = int(bspline_surface.get('v_degree')[()])
        self.v_knots = np.array(bspline_surface.get('v_knots')[()]).reshape(-1, 1).T
        self.v_rational = bool(bspline_surface.get('v_rational')[()])
        self.weights = np.column_stack(
            [bspline_surface['weights'][str(i)][()] for i in range(len(bspline_surface['weights']))]).T

        self.u_periodic = bool(bspline_surface.get('u_periodic')[()])
        self.v_periodic = bool(bspline_surface.get('v_periodic')[()])
        self.transform = np.array(bspline_surface.get('transform')[()])
        self.area = -1
        self.shape_name = bspline_surface.get('type')[()].decode('utf8')

        if self.u_rational or self.v_rational:
            self.surface_obj = NURBS.Surface(normalize_kv=False)
        else:
            self.surface_obj = BSpline.Surface(normalize_kv=False)

        self.surface_obj.degree_u = self.u_degree
        self.surface_obj.degree_v = self.v_degree
        self.surface_obj.ctrlpts_size_u = self.poles.shape[0]
        self.surface_obj.ctrlpts_size_v = self.poles.shape[1]
        self.surface_obj.knotvector_u = self.u_knots.squeeze().tolist()
        self.surface_obj.knotvector_v = self.v_knots.squeeze().tolist()
        self.surface_obj.ctrlpts = self.poles.reshape(-1, 3).tolist()
        if self.u_rational or self.v_rational:
            self.surface_obj.weights = self.weights.flatten().tolist()



    def sample(self, sample_points):
        if sample_points.size == 0:
            return self.poles[0, 0]
        uv_pairs = [(sample_points[i, 0], sample_points[i, 1]) for i in range(len(sample_points[:, 0]))]
        return np.array(self.surface_obj.evaluate_list(uv_pairs))

    def derivative(self, sample_points, order=1):
        if order == 0:
            return self.sample(sample_points)
        elif order == 1:
            res = np.zeros((sample_points.shape[0], 3, 2))
            for i in range(sample_points.shape[0]):
                d = self.surface_obj.derivatives(sample_points[i, 0], sample_points[i, 1], order)
                res[i, :, 0] = d[1][0]
                res[i, :, 1] = d[0][1]

            return res
        elif order == 2:
            res = np.zeros((sample_points.shape[0], 3, 2, 2))
            for i in range(sample_points.shape[0]):
                d = self.surface_obj.derivatives(sample_points[i, 0], sample_points[i, 1], order)
                res[i, :, 0, 0] = d[2][0]
                res[i, :, 1, 0] = d[1][1]
                res[i, :, 0, 1] = d[1][1]
                res[i, :, 1, 1] = d[0][2]

            return res
        else:
            raise ValueError("Order must be 0, 1, or 2")

    def normal(self, sample_points):
        if sample_points.size == 0:
            return np.array([])
        normals = operations.normal(self.surface_obj, sample_points)
        normal_vectors = np.array([n[-1] for n in normals])

        return normal_vectors



class Extrusion(Surface):
    def __init__(self, extrusion):
        self.direction = np.array(extrusion.get('direction')[()]).reshape(-1, 1).T
        self.trim_domain = np.array(extrusion.get('trim_domain')[()])
        self.transform = np.array(extrusion.get('transform')[()])
        self.area = -1
        self.shape_name = extrusion.get('type')[()].decode('utf8')
        _, self.curve = create_curve(extrusion['curve'], False)

    def sample(self, points):
        return self.curve.sample(points[:, 0][:, np.newaxis]) + points[:, 1][:, np.newaxis] * self.direction

    def derivative(self, points, order=1):
        if order == 0:
            return self.sample(points)
        elif order == 1:
            res = np.zeros((points.shape[0], 3, 2))
            res[:, :, 0] = self.curve.derivative(points[:, 0][:, np.newaxis], order=1)
            res[:, :, 1] = self.direction
            return res
        elif order == 2:
            res = np.zeros((points.shape[0], 3, 2, 2))
            res[:, :, 0, 0] = self.curve.derivative(points[:, 0][:, np.newaxis], order=2)

        return res


class Revolution(Surface):
    def __init__(self, revolution):
        self.location = np.array(revolution.get('location')[()]).reshape(-1, 1).T
        self.trim_domain = np.array(revolution.get('trim_domain')[()])
        self.transform = np.array(revolution.get('transform')[()])
        self.area = -1
        self.shape_name = revolution.get('type')[()].decode('utf8')
        self.z_axis = np.array(revolution.get('z_axis')[()]).reshape(-1, 1).T
        _, self.curve = create_curve(revolution['curve'], False)

    def sample(self, points):
        dx, dy, dz = self.z_axis[0, 0], self.z_axis[0, 1], self.z_axis[0, 2]

        cos_t = np.cos(points[:, 0])
        sin_t = np.sin(points[:, 0])
        one_minus_cos = 1 - cos_t

        R = np.array([
            [cos_t + dx * dx * one_minus_cos, dx * dy * one_minus_cos - dz * sin_t,
             dx * dz * one_minus_cos + dy * sin_t],
            [dy * dx * one_minus_cos + dz * sin_t, cos_t + dy * dy * one_minus_cos,
             dy * dz * one_minus_cos - dx * sin_t],
            [dz * dx * one_minus_cos - dy * sin_t, dz * dy * one_minus_cos + dx * sin_t,
             cos_t + dz * dz * one_minus_cos]
        ])

        return np.einsum('ijk,ik->ij', R.transpose(2, 0, 1), self.curve.sample(points[:, 1][:, np.newaxis]) - self.location) + self.location

    def derivative(self, points, order=1):
        if order == 0:
            return self.sample(points)
        elif order == 1:
            res = np.zeros((points.shape[0], 3, 2))
            theta = points[:, 0]
            u = self.z_axis[0, :]
            v_param = points[:, 1]
            dx, dy, dz = self.z_axis[0, 0], self.z_axis[0, 1], self.z_axis[0, 2]

            cos_t = np.cos(points[:, 0])
            sin_t = np.sin(points[:, 0])
            one_minus_cos = 1 - cos_t

            cos_t = np.cos(theta)
            sin_t = np.sin(theta)

            # outer product u ⊗ u, shape (3, 3)
            uuT = np.outer(u, u)
            I = np.eye(3)
            u_skew = np.array([
                [0, -u[2], u[1]],
                [u[2], 0, -u[0]],
                [-u[1], u[0], 0]
            ])

            # dR/dtheta, vectorized version for each theta
            dR_dtheta = (
                -sin_t[:, None, None] * I
                + sin_t[:, None, None] * uuT
                + cos_t[:, None, None] * u_skew
            )  # shape (N, 3, 3)

            R = np.array([
                [cos_t + dx * dx * one_minus_cos, dx * dy * one_minus_cos - dz * sin_t,
                 dx * dz * one_minus_cos + dy * sin_t],
                [dy * dx * one_minus_cos + dz * sin_t, cos_t + dy * dy * one_minus_cos,
                 dy * dz * one_minus_cos - dx * sin_t],
                [dz * dx * one_minus_cos - dy * sin_t, dz * dy * one_minus_cos + dx * sin_t,
                 cos_t + dz * dz * one_minus_cos]
            ])

            # v = sampled curve points - self.location
            v = self.curve.sample(v_param[:, None]) - self.location  # shape (N, 3)
            dv_dtheta = np.einsum('nij,nj->ni', dR_dtheta, v)

            # df/dv (i.e., derivative of curve)
            df_dv = self.curve.derivative(v_param[:, None], order=1)  # shape (N, 3)
            dv_dv = np.einsum('nij,nj->ni', R.transpose(2, 0, 1), df_dv)
            res[:, :, 0] = dv_dtheta
            res[:, :, 1] = dv_dv
        elif order == 2:
            res = np.zeros((points.shape[0], 3, 2, 2))

            u = self.z_axis[0,:]

            N = len(points)
            theta = points[:, 0]
            v_param = points[:, 1]

            cos_t = np.cos(theta)
            sin_t = np.sin(theta)
            one_minus_cos = 1 - cos_t

            # Basic components
            uuT = np.outer(u, u)
            I = np.eye(3)
            u_skew = skew(u)

            # Rotation matrices R
            R = (
                cos_t[:, None, None] * I
                + one_minus_cos[:, None, None] * uuT
                + sin_t[:, None, None] * u_skew
            )  # shape (N, 3, 3)

            # Derivative of R wrt theta
            dR_dtheta = (
                -sin_t[:, None, None] * I
                + sin_t[:, None, None] * uuT
                + cos_t[:, None, None] * u_skew
            )  # shape (N, 3, 3)

            # Second derivative of R wrt theta
            d2R_dtheta2 = (
                -cos_t[:, None, None] * I
                + cos_t[:, None, None] * uuT
                - sin_t[:, None, None] * u_skew
            )  # shape (N, 3, 3)

            # Sample and differentiate the curve
            f = self.curve.sample(v_param[:, None]) - self.location  # shape (N, 3)
            df_dv = self.curve.derivative(v_param[:, None], order=2)  # shape (N, 3)
            d2f_dv2 = self.curve.derivative(v_param[:, None], order=2)  # shape (N, 3)


            # Second derivative outputs
            d2y_dtheta2 = np.einsum('nij,nj->ni', d2R_dtheta2, f)
            d2y_dthetadv = np.einsum('nij,nj->ni', dR_dtheta, df_dv)
            d2y_dv2 = np.einsum('nij,nj->ni', R, d2f_dv2)

            res[:, :, 0, 0] = d2y_dtheta2
            res[:, :, 0, 1] = d2y_dthetadv
            res[:, :, 1, 0] = d2y_dthetadv
            res[:, :, 1, 1] = d2y_dv2

        return res


class Offset(Surface):
    def __init__(self, offset):
        self.trim_domain = np.array(offset.get('trim_domain')[()])
        self.transform = np.array(offset.get('transform')[()])
        self.area = -1
        self.value = np.float64(offset.get('value')[()])
        self.shape_name = offset.get('type')[()].decode('utf8')
        _, self.surface = create_surface(offset['surface'], False)

    def sample(self, points):

        N = points.shape[0]
        base_points = self.surface.sample(points)
        normals = self.surface.normal(points)
        offset_points = base_points + self.value * normals

        return offset_points

    def derivative(self, points, order=1):
        if order == 0:
            return self.sample(points)
        elif order == 1:
            diff = self.surface.derivative(points, order=1)

            res = diff.copy()

            du = diff[:, :, 0]
            dv = diff[:, :, 1]

            diffdiff = self.surface.derivative(points, order=2)

            c = np.cross(du, dv) # diff[:, :, 0] x diff[:, :, 1]
            norm = np.linalg.norm(c, axis=1)[:, None]
            dcdu = np.cross(diffdiff[:, :, 0, 0], diff[:, :, 1]) + np.cross(diff[:, :, 0], diffdiff[: ,:, 1, 0])
            dcdv = np.cross(diffdiff[:, :, 0, 1], diff[:, :, 1]) + np.cross(diff[:, :, 0], diffdiff[:, :, 1, 1])

            res[:, :, 0] += self.value * (dcdu/norm - c @ (c.T @ dcdu)/norm**3)
            res[:, :, 1] += self.value * (dcdv/norm - c @ (c.T @ dcdv)/norm**3)

            return res




from abs.utils import *
from abs.part_processor import *
import unittest
from abs.topology import Face
import os

def get_normal_func(part, topo, points):
    if isinstance(topo, Face):
        return topo.normal(points)
    else:
        return None





class TestShapeFunctions(unittest.TestCase):

    def test_get_parts_integration(self):
        name = 'Cone'
        sample_name = f'{name}.hdf5'
        file_path = os.path.join(os.path.dirname(__file__), '..', 'data', 'sample_hdf5', sample_name)
        file_path = os.path.normpath(file_path)

        num_samples = 5000

        parts = read_parts(file_path)

        P, S = sample_parts(parts, num_samples, get_normal_func)
        save_file_path = os.path.join(os.path.dirname(__file__), '..', 'test', 'sample_results', f'{name}_normals.obj')

        # save_ply(save_file_path, P, S)
        # save_obj(save_file_path, P)

        self.assertEqual(len(P), len(parts))
        self.assertEqual(len(S), len(parts))

        self.assertEqual(len(P), len(S))

if __name__ == '__main__':
    unittest.main()



import unittest
from random import sample
import numpy as np
from .test_utilities import *
from scipy.interpolate import BSpline

def surface_derivative(surface, sample_points, epsilon=1e-6):
    # testing ds/du order 1
    sample_points_plus = sample_points.copy()
    sample_points_plus = sample_points.copy()
    sample_points_plus[:, 0] += epsilon
    deriv = (surface.sample(sample_points_plus) - surface.sample(sample_points)) / epsilon
    deriv1 = surface.derivative(sample_points, 1)
    p = np.abs(deriv - deriv1[:, :, 0]).max()

    # testing ds/dv order 1
    sample_points_plus = sample_points.copy()
    sample_points_plus[:, 1] += epsilon
    deriv = (surface.sample(sample_points_plus) - surface.sample(sample_points)) / epsilon
    deriv1 = surface.derivative(sample_points, 1)
    q = np.abs(deriv - deriv1[:, :, 1]).max()

    # testing ds/du order 2
    sample_points_plus = sample_points.copy()
    sample_points_plus[:, 0] += epsilon
    deriv = (surface.derivative(sample_points_plus, 1) - surface.derivative(sample_points, 1)) / epsilon
    deriv1 = surface.derivative(sample_points, 2)
    p2 = np.abs(deriv - deriv1[:, :, :, 0]).max()

    # testing ds/dv order 2
    sample_points_plus = sample_points.copy()
    sample_points_plus[:, 1] += epsilon
    deriv = (surface.derivative(sample_points_plus, 1) - surface.derivative(sample_points, 1)) / epsilon
    deriv1 = surface.derivative(sample_points, 2)
    q2 = np.abs(deriv - deriv1[:, :, :, 1]).max()

    return p, q, p2, q2


def curves_derivative(curve, sample_points):
    epsilon = (sample_points[1] - sample_points[0]) * 1e-5

    sample_points_plus = sample_points.copy()
    sample_points_plus += epsilon
    deriv = (curve.sample(sample_points_plus) - curve.sample(sample_points)) / epsilon
    deriv1 = curve.derivative(sample_points, 1)
    p = np.abs(deriv - deriv1).max()/max(1, np.abs(np.mean(deriv)))

    sample_points_plus = sample_points.copy()
    sample_points_plus += epsilon
    deriv = (curve.derivative(sample_points_plus, 1) - curve.derivative(sample_points, 1)) / epsilon
    deriv1 = curve.derivative(sample_points, 2)
    q = np.abs(deriv - deriv1).max()/max(1, np.abs(np.mean(deriv)))

    return p, q


def generate_points_on_curve(curve, num_samples=1000):
    param_range = np.linspace(curve.interval[0, 0], curve.interval[0, 1], num_samples)
    param_range = param_range[:, None]
    points = curve.sample(param_range.reshape(-1, 1))
    return param_range, points


# def estimate_normal(curve, num_samples=1000):

#     _, points = generate_points_on_curve(curve, num_samples)

#     if points.shape[1] == 3:
#         tangents = np.gradient(points, axis=0)
#         tangents = tangents / np.linalg.norm(tangents, axis=1, keepdims=True)
#         dT = np.gradient(tangents, axis=0)
#         norms = np.linalg.norm(dT, axis=1, keepdims=True)
#         norms[norms == 0] = 1.0
#         normals = dT / norms
#         return normals
#     else:
#         lines = np.diff(points, axis=0)
#         lengths = np.linalg.norm(lines, axis=1)
#         normalized_lines = lines / lengths[:, np.newaxis]
#         rotation_matrix = np.array([[0, -1], [1, 0]])
#         rotated_p = normalized_lines @ rotation_matrix.T
#         return rotated_p


class TestGeometry(unittest.TestCase):

    def test_shapes(self):
        line3d()
        line2d()

        circle3d()
        circle2d()

        ellipse3d()
        ellipse2d()

        bspline_curve3d()
        bspline_curve2d()

        plane()
        cylinder()
        cone()
        sphere()
        torus()
        bspline_surface()

    def test_line2d(self):

        shape = line2d()
        self.assertEqual(shape.location.shape, (1, 2))
        self.assertEqual(shape.direction.shape, (1, 2))
        self.assertEqual(shape.interval.shape, (1, 2))
        sample_points = np.linspace(0, 1, 10).reshape(-1, 1)

        # sampling
        self.assertEqual(shape.sample(sample_points).shape, (10, 2))

        # derivative
        self.assertEqual(shape.derivative(sample_points, 0).shape, (10, 2))
        self.assertEqual(shape.derivative(sample_points, 1).shape, (10, 2))
        self.assertEqual(shape.derivative(sample_points, 2).shape, (10, 2))
        d, d2 = curves_derivative(shape, sample_points)
        self.assertTrue(d < 1e-7)
        self.assertTrue(d2 < 1e-7)

        # length
        num_samples = 1000
        param_points, points = generate_points_on_curve(shape, num_samples)
        self.assertTrue(abs(np.sum(np.linalg.norm(np.diff(points, axis=0), axis=1)) -
                            (shape.length if shape.length != -1 else shape.get_length()) < 1e-4))

        # normals
        # rotated_p = estimate_normal(shape, num_samples)
        # self.assertTrue(abs(np.sum(rotated_p - shape.normal(param_points)[1:, :]) < 1e-4))

        # check if normals are unit length
        # self.assertTrue(np.allclose(np.linalg.norm(shape.normal(param_points), axis=1), 1, atol=1e-8))

    def test_circle2d(self):
        shape = circle2d()
        self.assertEqual(shape.location.shape, (1, 2))
        self.assertEqual(type(shape.radius), float)
        self.assertEqual(shape.interval.shape, (1, 2))
        self.assertEqual(shape.x_axis.shape, (1, 2))
        self.assertEqual(shape.y_axis.shape, (1, 2))

        # sampling
        sample_points = np.linspace(0, 2 * np.pi, 10).reshape(-1, 1)
        self.assertEqual(shape.sample(sample_points).shape, (10, 2))

        # derivative
        self.assertEqual(shape.derivative(sample_points, 0).shape, (10, 2))
        self.assertEqual(shape.derivative(sample_points, 1).shape, (10, 2))
        self.assertEqual(shape.derivative(sample_points, 2).shape, (10, 2))
        self.assertEqual(shape.derivative(sample_points, 3).shape, (10, 2))
        self.assertEqual(shape.derivative(sample_points, 4).shape, (10, 2))
        d, d2 = curves_derivative(shape, sample_points)
        self.assertTrue(d < 1e-4)
        self.assertTrue(d2 < 1e-4)

        # length
        num_samples = 1000
        param_points, points = generate_points_on_curve(shape, num_samples)
        self.assertTrue(abs(np.sum(np.linalg.norm(np.diff(points, axis=0), axis=1)) -
                            (shape.length if shape.length != -1 else shape.get_length()) < 1e-4))

        # normals
        # rotated_p = estimate_normal(shape, num_samples)
        # self.assertTrue(abs(np.sum(rotated_p - shape.normal(param_points)[1:, :]) < 1e-4))

        # check if normals are unit length
        # self.assertTrue(np.allclose(np.linalg.norm(shape.normal(param_points), axis=1), 1, atol=1e-8))

    def test_ellipse2d(self):

        ellipse = ellipse2d()
        self.assertEqual(ellipse.focus1.shape, (1, 2))
        self.assertEqual(ellipse.focus2.shape, (1, 2))
        self.assertEqual(ellipse.interval.shape, (1, 2))
        self.assertEqual(type(ellipse.maj_radius), float)
        self.assertEqual(type(ellipse.min_radius), float)
        self.assertEqual(ellipse.x_axis.shape, (1, 2))
        self.assertEqual(ellipse.y_axis.shape, (1, 2))
        sample_points = np.linspace(0, 2 * np.pi, 10).reshape(-1, 1)
        self.assertEqual(ellipse.sample(sample_points).shape, (10, 2))

        # derivative
        self.assertEqual(ellipse.derivative(sample_points, 0).shape, (10, 2))
        self.assertEqual(ellipse.derivative(sample_points, 1).shape, (10, 2))
        self.assertEqual(ellipse.derivative(sample_points, 2).shape, (10, 2))
        self.assertEqual(ellipse.derivative(sample_points, 3).shape, (10, 2))
        self.assertEqual(ellipse.derivative(sample_points, 4).shape, (10, 2))
        d, d2 = curves_derivative(ellipse, sample_points)
        self.assertTrue(d < 1e-4)
        self.assertTrue(d2 < 1e-4)

        # length
        num_samples = 1000  # Can be adjusted for precision
        param_points, points = generate_points_on_curve(ellipse, num_samples)
        self.assertTrue(abs(np.sum(np.linalg.norm(np.diff(points, axis=0), axis=1)) -
                            (ellipse.length if ellipse.length != -1 else ellipse.get_length()) < 1e-4))

        # normals
        # rotated_p = estimate_normal(ellipse, num_samples)
        # self.assertTrue(abs(np.sum(rotated_p - ellipse.normal(param_points)[1:, :]) < 1e-4))

        # check if normals are unit length
        # self.assertTrue(np.allclose(np.linalg.norm(ellipse.normal(param_points), axis=1), 1, atol=1e-8))

    def test_bspline_curve2d(self):

        shape = bspline_curve2d()

        self.assertEqual(shape.closed, False)
        self.assertEqual(type(shape.continuity), int)
        self.assertEqual(type(shape.degree), int)
        self.assertEqual(shape.interval.shape, (1, 2))
        self.assertEqual(shape.knots.shape[0], 1)
        self.assertEqual(shape.poles.shape[1], 2)
        self.assertEqual(type(shape.rational), bool)
        self.assertEqual(shape.weights.shape[1], 1)
        self.assertEqual(shape.poles.shape[0], shape.weights.shape[0])

        # sample points
        umin_value, umax_value = shape.interval.T
        gridX = np.linspace(umin_value, umax_value)
        sample_points = gridX.reshape(-1, 1)

        self.assertEqual(shape.sample(sample_points).shape, (gridX.shape[0], 2))
        self.assertEqual(shape.derivative(sample_points, 0).shape, (50, 2))
        self.assertEqual(shape.derivative(sample_points, 1).shape, (50, 2))

        d, d2 = curves_derivative(shape, sample_points)

        self.assertTrue(d < 1e-4)
        self.assertTrue(d2 < 1e-4)

        # length
        num_samples = 1000
        param_points, points = generate_points_on_curve(shape, num_samples)
        self.assertTrue(abs(np.sum(np.linalg.norm(np.diff(points, axis=0), axis=1)) - shape.get_length() < 1e-4))

        # normals
        # rotated_p = estimate_normal(shape, num_samples)
        # self.assertTrue(abs(np.sum(rotated_p - shape.normal(param_points)[1:, :]) < 1e-4))

        # check if normals are unit length
        # self.assertTrue(np.allclose(np.linalg.norm(shape.normal(param_points), axis=1), 1, atol=1e-8))



    def test_line3d(self):
        shape = line3d()
        self.assertEqual(shape.location.shape, (1, 3))
        self.assertEqual(shape.direction.shape, (1, 3))
        self.assertEqual(shape.interval.shape, (1, 2))
        sample_points: None = np.linspace(0, 1, 10).reshape(-1, 1)

        # sampling
        self.assertEqual(shape.sample(sample_points).shape, (10, 3))

        # derivative shape
        self.assertEqual(shape.derivative(sample_points, 0).shape, (10, 3))
        self.assertEqual(shape.derivative(sample_points, 1).shape, (10, 3))
        self.assertEqual(shape.derivative(sample_points, 2).shape, (10, 3))

        d, d2 = curves_derivative(shape, sample_points)
        self.assertTrue(d < 1e-7)
        self.assertTrue(d2 < 1e-7)

        # length
        num_samples = 1000
        param_points, points = generate_points_on_curve(shape, num_samples)
        self.assertTrue(abs(np.sum(np.linalg.norm(np.diff(points, axis=0), axis=1)) - shape.get_length() < 1e-4))

    def test_circle3d(self):
        shape = circle3d()
        self.assertEqual(shape.location.shape, (1, 3))
        self.assertEqual(type(shape.radius), float)
        self.assertEqual(shape.interval.shape, (1, 2))
        self.assertEqual(shape.x_axis.shape, (1, 3))
        self.assertEqual(shape.y_axis.shape, (1, 3))
        self.assertEqual(shape.z_axis.shape, (1, 3))
        sample_points = np.linspace(0, 2 * np.pi, 10).reshape(-1, 1)
        self.assertEqual(shape.sample(sample_points).shape, (10, 3))

        # derivative
        self.assertEqual(shape.derivative(sample_points, 0).shape, (10, 3))
        self.assertEqual(shape.derivative(sample_points, 1).shape, (10, 3))
        self.assertEqual(shape.derivative(sample_points, 2).shape, (10, 3))
        self.assertEqual(shape.derivative(sample_points, 3).shape, (10, 3))
        self.assertEqual(shape.derivative(sample_points, 4).shape, (10, 3))

        d, d2 = curves_derivative(shape, sample_points)
        self.assertTrue(d < 1e-4)
        self.assertTrue(d2 < 1e-4)

        # length
        num_samples = 1000
        param_points, points = generate_points_on_curve(shape, num_samples)
        self.assertTrue(abs(np.sum(np.linalg.norm(np.diff(points, axis=0), axis=1)) - shape.get_length() < 1e-4))

    def test_ellipse3d(self):
        ellipse = ellipse3d()
        self.assertEqual(ellipse.focus1.shape, (1, 3))
        self.assertEqual(ellipse.focus2.shape, (1, 3))
        self.assertEqual(ellipse.interval.shape, (1, 2))
        self.assertEqual(type(ellipse.maj_radius), float)
        self.assertEqual(type(ellipse.min_radius), float)
        self.assertEqual(ellipse.x_axis.shape, (1, 3))
        self.assertEqual(ellipse.y_axis.shape, (1, 3))
        self.assertEqual(ellipse.z_axis.shape, (1, 3))
        sample_points = np.linspace(0, 2 * np.pi, 10).reshape(-1, 1)
        self.assertEqual(ellipse.sample(sample_points).shape, (10, 3))
        # derivative
        self.assertEqual(ellipse.derivative(sample_points, 0).shape, (10, 3))
        self.assertEqual(ellipse.derivative(sample_points, 1).shape, (10, 3))
        self.assertEqual(ellipse.derivative(sample_points, 2).shape, (10, 3))
        self.assertEqual(ellipse.derivative(sample_points, 3).shape, (10, 3))
        self.assertEqual(ellipse.derivative(sample_points, 4).shape, (10, 3))

        d, d2 = curves_derivative(ellipse, sample_points)
        self.assertTrue(d < 1e-4)
        self.assertTrue(d2 < 1e-4)

        # length
        num_samples = 1000
        param_points, points = generate_points_on_curve(ellipse, num_samples)
        self.assertTrue(abs(np.sum(np.linalg.norm(np.diff(points, axis=0), axis=1)) - ellipse.get_length() < 1e-4))

        #normals
        # rotated_p = estimate_normal(ellipse, num_samples)

    def test_bspline_curve3d(self):
        shape = bspline_curve3d()
        self.assertEqual(type(shape.closed), bool)
        self.assertEqual(type(shape.continuity), int)
        self.assertEqual(type(shape.degree), int)
        self.assertEqual(shape.interval.shape, (1, 2))
        self.assertEqual(shape.knots.shape[0], 1)
        self.assertEqual(shape.poles.shape[1], 3)
        self.assertEqual(type(shape.rational), bool)
        self.assertEqual(shape.weights.shape[1], 1)

        # sample points
        umin_value, umax_value = shape.interval.T
        gridX = np.linspace(umin_value, umax_value)
        sample_points = gridX.reshape(-1, 1)
        self.assertEqual(shape.sample(sample_points).shape, (gridX.shape[0], 3))

        # derivative
        self.assertEqual(shape.derivative(sample_points, 0).shape, (50, 3))
        self.assertEqual(shape.derivative(sample_points, 1).shape, (50, 3))
        d, d2 = curves_derivative(shape, sample_points)
        self.assertTrue(d < 1e-4)
        self.assertTrue(d2 < 1e-4)

        # length
        num_samples = 1000
        param_points, points = generate_points_on_curve(shape, num_samples)
        self.assertTrue(abs(np.sum(np.linalg.norm(np.diff(points, axis=0), axis=1)) - shape.get_length() < 1e-4))


    def test_plane(self):
        shape = plane()
        self.assertEqual(shape.coefficients.shape, (1, 4))
        self.assertEqual(shape.location.shape, (1, 3))
        self.assertEqual(shape.trim_domain.shape, (2, 2))
        self.assertEqual(shape.x_axis.shape, (1, 3))
        self.assertEqual(shape.y_axis.shape, (1, 3))
        self.assertEqual(shape.z_axis.shape, (1, 3))

        # sample points
        umin_value, umax_value, vmin_value, vmax_value = shape.trim_domain.reshape(-1, 1)
        gridX = np.linspace(umin_value, umax_value)
        gridY = np.linspace(vmin_value, vmax_value)
        gridX, gridY = np.meshgrid(gridX, gridY)
        gridX = gridX.reshape((np.prod(gridX.shape),)).reshape(-1, 1)
        gridY = gridY.reshape((np.prod(gridY.shape),)).reshape(-1, 1)
        sample_points = np.column_stack((gridX, gridY)).reshape(-1, 2)

        self.assertEqual(shape.sample(sample_points).shape, (gridX.shape[0] * gridX.shape[1], 3))

        # derivative:
        p, q, du, dv = surface_derivative(shape, sample_points)
        self.assertTrue(p < 1e-7)
        self.assertTrue(q < 1e-7)
        self.assertTrue(du < 1e-7)
        self.assertTrue(dv < 1e-7)

        # self.assertEqual(int(shape.get_area()), 1)
        self.assertTrue(abs(shape.get_area() - 1) < 1e-7)

    def test_cylinder(self):
        shape = cylinder()
        self.assertEqual(shape.location.shape, (1, 3))
        self.assertEqual(type(shape.radius), float)
        self.assertEqual(shape.coefficients.shape, (1, 10))
        self.assertEqual(shape.trim_domain.shape, (2, 2))
        self.assertEqual(shape.x_axis.shape, (1, 3))
        self.assertEqual(shape.y_axis.shape, (1, 3))
        self.assertEqual(shape.z_axis.shape, (1, 3))

        # sample points
        umin_value, umax_value, vmin_value, vmax_value = shape.trim_domain.reshape(-1, 1)
        gridX = np.linspace(umin_value, umax_value)
        gridY = np.linspace(vmin_value, vmax_value)
        gridX, gridY = np.meshgrid(gridX, gridY)
        sample_points = np.column_stack((gridX, gridY)).reshape(-1, 2)
        self.assertEqual(shape.sample(sample_points).shape, (gridX.shape[0] * gridX.shape[1], 3))

        du, dv, d2u, d2v = surface_derivative(shape, sample_points)
        self.assertTrue(du < 1e-6)
        self.assertTrue(dv < 1e-6)
        self.assertTrue(d2u < 1e-6)
        self.assertTrue(d2v < 1e-6)

        self.assertTrue(abs(shape.get_area() - 1) < 1e-7)

    def test_cone(self):
        shape = cone()
        self.assertEqual(type(shape.angle), float)
        self.assertEqual(shape.apex.shape, (1, 3))
        self.assertEqual(shape.coefficients.shape, (1, 10))
        self.assertEqual(shape.location.shape, (1, 3))
        self.assertEqual(type(shape.radius), float)
        self.assertEqual(shape.trim_domain.shape, (2, 2))
        self.assertEqual(shape.x_axis.shape, (1, 3))
        self.assertEqual(shape.y_axis.shape, (1, 3))
        self.assertEqual(shape.z_axis.shape, (1, 3))

        # sample points
        umin_value, umax_value, vmin_value, vmax_value = shape.trim_domain.reshape(-1, 1)
        gridX = np.linspace(umin_value, umax_value)
        gridY = np.linspace(vmin_value, vmax_value)
        gridX, gridY = np.meshgrid(gridX, gridY)
        sample_points = np.column_stack((gridX, gridY)).reshape(-1, 2)
        self.assertEqual(shape.sample(sample_points).shape, (gridX.shape[0] * gridX.shape[1], 3))

        du, dv, d2u, d2v = surface_derivative(shape, sample_points)
        self.assertTrue(du < 1e-6)
        self.assertTrue(dv < 1e-6)
        self.assertTrue(d2u < 1e-6)
        self.assertTrue(d2v < 1e-6)


    def test_sphere(self):
        shape = sphere()
        self.assertEqual(shape.coefficients.shape, (1, 10))
        self.assertEqual(shape.location.shape, (1, 3))
        self.assertEqual(type(shape.radius), float)
        self.assertEqual(shape.trim_domain.shape, (2, 2))
        self.assertEqual(shape.x_axis.shape, (1, 3))
        self.assertEqual(shape.y_axis.shape, (1, 3))
        self.assertEqual(shape.z_axis.shape, (1, 3))

        # sample points
        umin_value, umax_value, vmin_value, vmax_value = shape.trim_domain.reshape(-1, 1)
        gridX = np.linspace(umin_value, umax_value)
        gridY = np.linspace(vmin_value, vmax_value)
        gridX, gridY = np.meshgrid(gridX, gridY)
        sample_points = np.column_stack((gridX, gridY)).reshape(-1, 2)
        self.assertEqual(shape.sample(sample_points).shape, (gridX.shape[0] * gridX.shape[1], 3))

        # derivative: du, dv
        du, dv, d2u, d2v = surface_derivative(shape, sample_points)
        self.assertTrue(du < 1e-6)
        self.assertTrue(dv < 1e-6)
        self.assertTrue(d2u < 1e-6)
        self.assertTrue(d2v < 1e-6)

    def test_torus(self):
        shape = torus()
        self.assertEqual(shape.location.shape, (1, 3))
        self.assertEqual(type(shape.max_radius), float)
        self.assertEqual(type(shape.min_radius), float)
        self.assertEqual(shape.trim_domain.shape, (2, 2))
        self.assertEqual(shape.x_axis.shape, (1, 3))
        self.assertEqual(shape.y_axis.shape, (1, 3))
        self.assertEqual(shape.z_axis.shape, (1, 3))

        # sample points
        umin_value, umax_value, vmin_value, vmax_value = shape.trim_domain.reshape(-1, 1)
        gridX = np.linspace(umin_value, umax_value)
        gridY = np.linspace(vmin_value, vmax_value)
        gridX, gridY = np.meshgrid(gridX, gridY)
        sample_points = np.column_stack((gridX, gridY)).reshape(-1, 2)
        self.assertEqual(shape.sample(sample_points).shape, (gridX.shape[0] * gridX.shape[1], 3))

        du, dv, d2u, d2v = surface_derivative(shape, sample_points)
        self.assertTrue(du < 1e-6)
        self.assertTrue(dv < 1e-6)
        self.assertTrue(d2u < 1e-6)
        self.assertTrue(d2v < 1e-6)

    def test_bspline_surface(self):
        shape = bspline_surface()

        self.assertEqual(type(shape.continuity), int)
        self.assertEqual(shape.face_domain.shape, (1, 4))
        self.assertEqual(type(shape.is_trimmed), bool)
        self.assertEqual(shape.poles.shape[2], 3)
        self.assertEqual(shape.trim_domain.shape, (2, 2))
        self.assertEqual(type(shape.u_closed), bool)
        self.assertEqual(type(shape.u_degree), int)
        self.assertEqual(shape.u_knots.shape[0], 1)
        self.assertEqual(type(shape.u_rational), bool)
        self.assertEqual(type(shape.v_closed), bool)
        self.assertEqual(type(shape.v_degree), int)
        self.assertEqual(shape.v_knots.shape[0], 1)
        self.assertEqual(type(shape.v_rational), bool)


        # sample points
        umin_value, umax_value, vmin_value, vmax_value = shape.trim_domain.reshape(-1, 1)
        gridX = np.linspace(umin_value, umax_value)
        gridY = np.linspace(vmin_value, vmax_value)
        gridX, gridY = np.meshgrid(gridX, gridY)
        sample_points = np.column_stack((gridX, gridY)).reshape(-1, 2)
        self.assertEqual(shape.sample(sample_points).shape, (gridX.shape[0] * gridX.shape[1], 3))

        # derivative
        du, dv, d2u, d2v = surface_derivative(shape, sample_points)
        self.assertTrue(du < 1e-6)
        self.assertTrue(dv < 1e-6)
        self.assertTrue(d2u < 1e-6)
        self.assertTrue(d2v < 1e-6)



if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""The setup script."""

import io
import re
import os
import sys
import subprocess
from pathlib import Path

from os import path as op
from setuptools import Extension, setup, find_packages
from setuptools.command.build_ext import build_ext



class CMakeExtension(Extension):
    def __init__(self, name: str, sourcedir: str = "") -> None:
        super().__init__(name, sources=[])
        self.sourcedir = os.fspath(Path(sourcedir).resolve())


class CMakeBuild(build_ext):
    def build_extension(self, ext: CMakeExtension) -> None:
        # Must be in this form due to bug in .resolve() only fixed in Python 3.10+
        ext_fullpath = Path.cwd() / self.get_ext_fullpath(ext.name)
        extdir = ext_fullpath.parent.resolve()

        # Using this requires trailing slash for auto-detection & inclusion of
        # auxiliary "native" libs

        debug = int(os.environ.get("DEBUG", 0)) if self.debug is None else self.debug
        cfg = "Debug" if debug else "Release"

        # CMake lets you override the generator - we need to check this.
        # Can be set with Conda-Build, for example.
        cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

        # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
        # EXAMPLE_VERSION_INFO shows you how to pass a value into the C++ code
        # from Python.
        cmake_args = [
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}{os.sep}",
            f"-DPYTHON_EXECUTABLE={sys.executable}",
            f"-DCMAKE_BUILD_TYPE={cfg}",  # not used on MSVC, but no harm
        ]
        build_args = []
        # Adding CMake arguments set as environment variable
        # (needed e.g. to build for ARM OSx on conda-forge)
        if "CMAKE_ARGS" in os.environ:
            cmake_args += [item for item in os.environ["CMAKE_ARGS"].split(" ") if item]

        # In this example, we pass in the version to C++. You might not need to.
        cmake_args += [f"-DEXAMPLE_VERSION_INFO={self.distribution.get_version()}"]

        if self.compiler.compiler_type != "msvc":
            # Using Ninja-build since it a) is available as a wheel and b)
            # multithreads automatically. MSVC would require all variables be
            # exported for Ninja to pick it up, which is a little tricky to do.
            # Users can override the generator with CMAKE_GENERATOR in CMake
            # 3.15+.
            if not cmake_generator or cmake_generator == "Ninja":
                try:
                    import ninja

                    ninja_executable_path = Path(ninja.BIN_DIR) / "ninja"
                    cmake_args += [
                        "-GNinja",
                        f"-DCMAKE_MAKE_PROGRAM:FILEPATH={ninja_executable_path}",
                    ]
                except ImportError:
                    pass

        else:
            # Single config generators are handled "normally"
            single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})

            # CMake allows an arch-in-generator style for backward compatibility
            contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})

            # Specify the arch if using MSVC generator, but only if it doesn't
            # contain a backward-compatibility arch spec already in the
            # generator name.
            # if not single_config and not contains_arch:
            #     cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

            # Multi-config generators have a different way to specify configs
            if not single_config:
                cmake_args += [
                    f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{cfg.upper()}={extdir}"
                ]
                build_args += ["--config", cfg]

        if sys.platform.startswith("darwin"):
            # Cross-compile support for macOS - respect ARCHFLAGS if set
            archs = re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", ""))
            if archs:
                cmake_args += ["-DCMAKE_OSX_ARCHITECTURES={}".format(";".join(archs))]

        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
        # across all generators.
        if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
            # self.parallel is a Python 3 only way to set parallel jobs by hand
            # using -j in the build_ext call, not supported by pip or PyPA-build.
            if hasattr(self, "parallel") and self.parallel:
                # CMake 3.12+ only.
                build_args += [f"-j{self.parallel}"]

        build_temp = Path(self.build_temp) / ext.name
        if not build_temp.exists():
            build_temp.mkdir(parents=True)

        subprocess.run(
            ["cmake", ext.sourcedir, *cmake_args], cwd=build_temp, check=True
        )
        subprocess.run(
            ["cmake", "--build", ".", *build_args], cwd=build_temp, check=True
        )


with open('README.md') as readme_file:
    readme = readme_file.read()

here = op.abspath(op.dirname(__file__))

# get the dependencies and installs
with io.open(op.join(here, "requirements.txt"), encoding="utf-8") as f:
    all_reqs = f.read().split("\n")

install_requires = [x.strip() for x in all_reqs if "git+" not in x]
dependency_links = [x.strip().replace("git+", "") for x in all_reqs if "git+" not in x]

requirements = [ ]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    python_requires='>=3.8',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    description="This project is a modular and extensible software framework designed for geometric processing, sampling, and topological analysis. It provides a comprehensive suite of tools for handling various geometric entities such as lines, circles, ellipses, and B-spline curves and surfaces",
    entry_points={
        'console_scripts': [
            'abs=abs.cli:main',
        ],
    },
    install_requires=install_requires,
    dependency_links=dependency_links,
    license="MIT license",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='abs',
    name='A Better Step',
    setup_requires=setup_requirements,
    test_suite='test',
    tests_require=test_requirements,
    ext_modules=[CMakeExtension('abspy')],
    cmdclass=dict(build_ext=CMakeBuild),
    url='https://better-step.github.io',
    version='0.0.1',
    zip_safe=False,
    packages=find_packages(include=['abs', 'abs.*', 'abspy'])
)



import unittest
from pathlib import Path
import h5py
import os
from abs.topology import Topology
from abs.utils import *
from .test_utilities import *

class TestTopology(unittest.TestCase):

    def test_topology(self):
        sample_name = 'Cone.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/topology']
            geo = hdf['parts/part_001/geometry']
            s = Shape(geo, grp)
            topo_solid = s.Solid


    def test_find_adjacent_faces(self):
        sample_name = 'Cone.hdf5'
        file_path = get_file(sample_name)
        with h5py.File(file_path, 'r') as hdf:
            grp = hdf['parts/part_001/topology']
            geo = hdf['parts/part_001/geometry']
            s = Shape(geo, grp)

        face_index = 1
        # adjacent_faces = s.Solid.faces[face_index]
        current_face = s.Solid.faces[face_index]
        adjacent_faces = current_face.find_adjacent_faces()
        # print(adjacent_faces)

    # def test_find_connected_components(self):
    #     sample_name = 'Cone.hdf5'
    #     file_path = get_file(sample_name)
    #     with h5py.File(file_path, 'r') as hdf:
    #         grp = hdf['parts/part_001/topology']
    #         geo = hdf['parts/part_001/geometry']
    #         s = Shape(geo, grp)
    #
    #     connected_components = topology.find_connected_components()
    #     self.assertEqual(len(connected_components[0]), 7)


    # def find_connected_components(self):
    #     """
    #     Identify connected components in the topology.
    #     """
    #     visited = set()
    #     components = []
    #
    #     def dfs(face_index):
    #         if face_index in visited:
    #             return
    #         visited.add(face_index)
    #         component.add(face_index)
    #         for adjacent_face in self.find_adjacent_faces(face_index):
    #             dfs(adjacent_face)
    #
    #     for face_index in range(len(self.faces)):
    #         if face_index not in visited:
    #             component = set()
    #             dfs(face_index)
    #             components.append(component)
    #
    #     return [list(component) for component in components]
if __name__ == '__main__':
    unittest.main()



from abs.part_processor import *
from abs.utils import *
import unittest
from test_utilities import *


def get_normal_func(part, topo, points):
    if topo.is_face():
        return topo.normal(points)
    else:
        return None


class TestApplication(unittest.TestCase):
    def test_run_application(self):
        file_path = get_file("Ellipse.hdf5")
        num_samples = 4000

        # print('getting parts')

        parts = read_parts(file_path)
        meshes = read_meshes(file_path)


        # print('sampling parts with sample size:', num_samples)
        P, S = sample_parts(parts, num_samples, get_normal_func)
        V, F = get_mesh(meshes)

        # print('saving results')
        # save_file_path = os.path.join(base_dir, f"{file_name}.obj")
        # save_file_path_ply = os.path.join(base_dir, f"{file_name}.ply")

        # save_ply(save_file_path_ply, P, S)
        # save_obj_mesh(save_file_path, V, F)

        # print('finished')